# Third party
import numpy as np
import numpy.matlib
from osgeo import gdal, osr
import rasterio
from scipy.spatial.transform import Rotation as RotLib
from scipy.spatial.transform import Slerp
import open3d as o3d
import xmltodict
from pyproj import CRS, Transformer
import pymap3d as pm
import ephem
import pandas as pd
from scipy.interpolate import interp1d
import pyvista as pv
from shapely.geometry import Polygon, mapping, MultiPoint
from rasterio.transform import from_origin
from rasterio.windows import from_bounds
from rasterio.windows import Window
import trimesh

# Python standard lib
import os
import time
from datetime import datetime
from dateutil import parser
import json

# A file were we define geometry and geometric transforms
class CalibHSI:
    def __init__(self, file_name_cal_xml, mode = 'r', param_dict = None):
        """
        :param file_name_cal_xml: str
        File name of calibration file for line camera model
        :param config: config
        global configuration object.
        :param mode: str
        open file for reading (for general use) or writing (post calibration)
        :param param_dict: dictionary:
        dictionary with keys
        "'rx', 'ry', 'rz', 'tx', 'ty', 'tz', 'cx', 'f', 'k1', 'k2', 'k3', 'width'"
        """
        if mode == 'r':
            with open(file_name_cal_xml, 'r', encoding='utf-8') as file:
                my_xml = file.read()
            xml_dict = xmltodict.parse(my_xml)
            self.calibrationHSI = xml_dict['calibration']

            self.w = float(self.calibrationHSI['width'])
            self.f = float(self.calibrationHSI['f'])
            self.cx = float(self.calibrationHSI['cx'])

            # Rotations
            self.rx = float(self.calibrationHSI['rx'])
            self.ry = float(self.calibrationHSI['ry'])
            self.rz = float(self.calibrationHSI['rz'])


            # Translations
            self.tx = float(self.calibrationHSI['tx'])
            self.ty = float(self.calibrationHSI['ty'])
            self.tz = float(self.calibrationHSI['tz'])

            # Distortions
            self.k1 = float(self.calibrationHSI['k1'])
            self.k2 = float(self.calibrationHSI['k2'])
            self.k3 = float(self.calibrationHSI['k3'])
        elif mode == 'w':
            # Check if the file exists
            if os.path.exists(file_name_cal_xml):
                with open(file_name_cal_xml, 'r', encoding='utf-8') as file:
                    my_xml = file.read()
                xml_dict = xmltodict.parse(my_xml)
            else:
                # Create a new xml_dict with default structure
                xml_dict = {'calibration': {}}

            # Update xml_dict['calibration'] with values from param_dict
            for key, value in param_dict.items():
                xml_dict['calibration'][key] = value
            with open(file_name_cal_xml, 'w') as fd:
                fd.write(xmltodict.unparse(xml_dict))

class CameraGeometry():
    def __init__(self, pos, rot, time, is_interpolated = False):
        
        self.position_nav = pos
        self.rotation_nav = rot

        self.time = time
        self.IsLocal = False
        self.decoupled = True

        if is_interpolated:
            self.position_nav_interpolated = self.position_nav
            self.rotation_nav_interpolated = self.rotation_nav



    def interpolate(self, time_hsi, minIndRGB, maxIndRGB, extrapolate):
        """"""
        # A simple interpolation of transforms where all images should be aligned.
        # Should also implement a more sensor fusion-like Kalman filter implementation
        self.time_hsi = time_hsi
        if self.decoupled == True:

            if extrapolate == False:
                time_interpolation = time_hsi[minIndRGB:maxIndRGB]
                linearPositionInterpolator = interp1d(self.time, np.transpose(self.position_nav))
                self.position_nav_interpolated = np.transpose(linearPositionInterpolator(time_interpolation))
                linearSphericalInterpolator = Slerp(self.time, self.rotation_nav)
                self.rotation_nav_interpolated = linearSphericalInterpolator(time_interpolation)
            else:
                # Extrapolation of position:
                time_interpolation = time_hsi
                linearPositionInterpolator = interp1d(self.time, np.transpose(self.position_nav), fill_value='extrapolate')
                self.position_nav_interpolated = np.transpose(linearPositionInterpolator(time_interpolation))

                # Extrapolation of orientation/rotation:

                # Synthetizize additional frames
                delta_rot_b1_b2 = (self.rotation_nav[-1].inv()) * (self.rotation_nav[-2])
                delta_time_last = self.time[-1] - self.time[-2]
                time_last = self.time[-1] + delta_time_last
                rot_last = self.rotation_nav[-1] * (delta_rot_b1_b2.inv()) # Assuming a continuation of the rotation

                # Rotation from second to first attitude
                delta_rot_b1_b2 = (self.rotation_nav[0].inv()) * (self.rotation_nav[1])
                delta_time_last = self.time[0] - self.time[1]
                time_first = self.time[0] + delta_time_last # Subtraction
                # Add the rotation from second to first attitude to the first attitude "continue" rotation
                rot_first = self.rotation_nav[0] * (delta_rot_b1_b2.inv())  # Assuming a continuation of the rotation
                time_concatenated = np.concatenate((np.array(time_first).reshape((1,-1)),
                                                    self.time.reshape((1,-1)),
                                                    np.array(time_last).reshape((1,-1))), axis = 1)\
                    .reshape(-1).astype(np.float64)
                rotation_list = [self.rotation_nav]
                rotation_list.append(rot_last)
                rotation_list.insert(0, rot_first)


                rot_vec_first = rot_first.as_rotvec().reshape((1,-1))
                rot_vec_mid = self.rotation_nav.as_rotvec()
                rot_vec_last = rot_last.as_rotvec().reshape((1,-1))

                rotation_vec_tot = np.concatenate((rot_vec_first, rot_vec_mid, rot_vec_last), axis = 0)

                Rotation_tot = RotLib.from_rotvec(rotation_vec_tot)



                time_concatenated = np.array(time_concatenated).astype(np.float64)

                linearSphericalInterpolator = Slerp(time_concatenated, Rotation_tot)

                self.rotation_nav_interpolated = linearSphericalInterpolator(time_interpolation)






        else:
            print('Proper interpolation of transformation with constant velocity and rotation has not yet been implemented')
            print('See https://www.geometrictools.com/Documentation/InterpolationRigidMotions.pdf')
            #self.rotation_nav_interpolated, self.PositionInterpolated = self.interpolateTransforms()
    def intrinsicTransformHSI(self, translation_ref_hsi, rot_hsi_ref_obj):

        # An intrinsic transform is a transformation to another reference frame on the moving body, i.e. the IMU or an RGB cam
        self.position_ecef = self.position_nav_interpolated + self.rotation_nav_interpolated.apply(translation_ref_hsi)

        # Composing rotations. See:
        # https: // docs.scipy.org / doc / scipy / reference / generated / scipy.spatial.transform.Rotation.__mul__.html
        self.rotation_hsi = self.rotation_nav_interpolated * rot_hsi_ref_obj

        self.quaternion_ecef = self.rotation_hsi.as_quat()
        
    def localTransform(self, frame):
        self.IsLocal = True
        self.LocalTransformFrame = frame

        if frame == 'ENU':
            self.position_nav_interpolated = self.Rotation_ecef_enu*self.position_nav_interpolated
            self.rotation_nav_interpolated = self.Rotation_ecef_enu*self.rotation_nav_interpolated
            self.position_nav = self.Rotation_ecef_enu * self.position_nav
            self.rotation_nav = self.Rotation_ecef_enu * self.rotation_nav
            self.rotation_hsi = self.Rotation_ecef_enu * self.rotation_hsi
            self.position_nav = self.Rotation_ecef_enu * self.position_ecef
        elif frame == 'NED':
            self.localPositionInterpolated = self.Rotation_ecef_ned*self.position_nav_interpolated
            self.localRotationInterpolated = self.Rotation_ecef_ned*self.rotation_nav_interpolated
            self.localPosition = self.Rotation_ecef_ned * self.position_nav
            self.localRotation = self.Rotation_ecef_ned * self.rotation_nav
            self.rotation_hsi = self.Rotation_ecef_ned * self.rotation_hsi
            self.position_nav = self.Rotation_ecef_ned * self.position_ecef
        else:
            print('Frame must be ENU or NED')
    def localTransformInverse(self):

        if self.IsLocal:
            self.IsLocal = False
            if self.LocalTransformFrame == 'ENU':
                self.position_nav_interpolated = self.Rotation_ecef_enu.inv()*self.position_nav_interpolated
                self.rotation_nav_interpolated = self.Rotation_ecef_enu.inv()*self.rotation_nav_interpolated
                self.position_nav = self.Rotation_ecef_enu.inv() * self.position_nav
                self.rotation_nav = self.Rotation_ecef_enu.inv() * self.rotation_nav
                self.rotation_hsi = self.Rotation_ecef_enu.inv() * self.rotation_hsi
                self.position_nav = self.Rotation_ecef_enu.inv() * self.position_ecef
            elif self.LocalTransformFrame == 'NED':
                self.localPositionInterpolated = self.Rotation_ecef_ned.inv()* self.position_nav_interpolated
                self.localRotationInterpolated = self.Rotation_ecef_ned.inv()*self.rotation_nav_interpolated
                self.localPosition = self.Rotation_ecef_ned.inv() * self.position_nav
                self.localRotation = self.Rotation_ecef_ned.inv() * self.rotation_nav
                self.rotation_hsi = self.Rotation_ecef_ned.inv() * self.rotation_hsi
                self.position_nav = self.Rotation_ecef_ned.inv() * self.position_ecef
            else:
                print('Frame must be ENU or NED')
        else:
            print('Poses are defined globally already')
    def defineRayDirections(self, dir_local):
        self.rayDirectionsLocal = dir_local

        n = self.position_ecef.shape[0]
        m = dir_local.shape[0]

        self.rayDirectionsGlobal = np.zeros((n, m, 3))

        # Converts data from local frame to global
        for i in range(n):
            self.rayDirectionsGlobal[i, :, :] = self.rotation_hsi[i].apply(dir_local)
    def intersect_with_mesh(self, mesh, max_ray_length, mesh_trans):
        """Intersects the rays of the camera with the 3D triangular mesh

        :param mesh: A mesh object read via the pyvista library
        :type mesh: Pyvista mesh
        :param max_ray_length: The upper bound length of the camera rays (it is determined )
        :type max_ray_length: _type_
        :param mesh_trans: The offset of the mesh
        :type mesh_trans: _type_
        """

        n = self.rayDirectionsGlobal.shape[0]
        m = self.rayDirectionsGlobal.shape[1]

        self.points_ecef_crs = np.zeros((n, m, 3), dtype=np.float64)
        self.normals_ecef_crs = np.zeros((n, m, 3), dtype=np.float64)

        # Duplicate multiple camera centres
        start_ECEF = np.einsum('ijk, ik -> ijk', np.ones((n, m, 3), dtype=np.float64), self.position_ecef).reshape((-1,3))

        # Subtract the mesh offset to avoid rounding errors and perform intersection tests with "small coordinates"
        start = start_ECEF - mesh_trans

        dir = (self.rayDirectionsGlobal * max_ray_length).reshape((-1,3))

        
        
        start_time = time.time()


        
        try:
            # This will only work if exact Python version is rigght and you have PyEmbree
            points, rays, cells = mesh.multi_ray_trace(origins=start, directions=dir, first_point=True, retry=True)
        except:
            # If you instead use embreex, python>3.6 will do
            
            
            # If the faces are 
            faces = mesh.regular_faces

            # Convert PolyData to trimesh.Trimesh
            tri_mesh = trimesh.Trimesh(vertices=mesh.points, faces=faces)

            # Define an intersector object
            ray_mesh_intersector = trimesh.ray.ray_pyembree.RayMeshIntersector(geometry=tri_mesh)

            # Intersect data
            count = 0
            cells, rays, points = ray_mesh_intersector.intersects_id(ray_origins=start,  
                                                                ray_directions=dir, 
                                                                multiple_hits=False,
                                                                return_locations=True)
            
        n_points = int(np.size(points)/3)
        n_rays = int(np.size(start)/3)
        # Recurring is that 1- 10s of rays fail to detect intersections using Trimesh.
        # A handy fix is to iterate the failed intersections using individual ray tracing
        # This equivalent to the retry=True in 
        # https://docs.pyvista.org/version/stable/api/core/_autosummary/pyvista.PolyDataFilters.multi_ray_trace.html#pyvista.PolyDataFilters.multi_ray_trace
        if n_points != n_rays:
            n_missing = n_rays - n_points
            print(f'Trimesh failed to find intersections for {n_missing} rays')
            print(f'Ray tracing for these rays is retried in pyvista')

            # Identify the missing intersections:
            missing_rays = np.array(list( set(range(n_rays)) - set(rays) ))

            # And ray trace them individually in VTK
            for ray in missing_rays:
                # Retry failed intersections with slow pyvista version
                point, cell = mesh.ray_trace(start[ray,:], start[ray,:] + dir[ray,:], first_point=True)
                cells = np.concatenate((cells, cell), axis=0)
                points = np.concatenate((points, point.reshape((1,3))), axis=0)
                rays = np.concatenate((rays, np.array([ray])), axis=0)
            n_points = int(np.size(points)/3)
            print('All rays successfully traced in VTK')

        else:
            print(f'All rays were successfully intersected with trimesh')
            


        stop_time = time.time()

        normals = mesh.cell_normals[cells,:]

        slit_image_number = np.floor(rays / m).astype(np.int32)

        pixel_number = rays % m

        


        # Assign normals
        self.points_ecef_crs[slit_image_number, pixel_number] = points + mesh_trans

        self.normals_ecef_crs[slit_image_number, pixel_number] = normals

        self.camera_to_seabed_ECEF = self.points_ecef_crs - start_ECEF.reshape((n, m, 3))


        # Calculate
        self.points_hsi_crs = np.zeros(self.camera_to_seabed_ECEF.shape)

        self.normals_hsi_crs = np.zeros(self.normals_ecef_crs.shape)

        self.depth_map = np.zeros((n, m))

        # For local geometry (when vehicle fixed artificial light is used):
        for i in range(n):
            # Calculate vector from HSI to seabed in local coordinates (for artificial illumination)
            self.points_hsi_crs[i, :, :] = (self.rotation_hsi[i].inv()).apply(self.camera_to_seabed_ECEF[i, :, :])

            # Calculate surface normals of intersected triangles (for artificial illumination)
            self.normals_hsi_crs[i,:,:] = (self.rotation_hsi[i].inv()).apply(self.normals_ecef_crs[i, :, :])

            # Calculate a depth map (the z-component, 1D scanline)
            self.depth_map[i, :] = self.points_hsi_crs[i, :, 2]/self.rayDirectionsLocal[:, 2]

        

        self.unix_time_grid = np.einsum('ijk, ik -> ijk', np.ones((n, m, 1), dtype=np.float64), self.time.reshape((-1,1)))
        self.unix_time = self.unix_time_grid.reshape((-1, 1)) # Vector-form
        self.pixel_nr_grid = np.matlib.repmat(np.arange(m), n, 1)
        self.frame_nr_grid = np.matlib.repmat(np.arange(n).reshape(-1,1), 1, m)

        
    @staticmethod
    def intersect_ray_with_earth_ellipsoid(p0, dir_hat, B):
        """_summary_

        :param p0: Ray origin
        :type p0: (3,1) array of floats 
        :param dir_hat: describing ray direction in some ECEF
        :type dir_hat: (3,1) array of floats 
        :param B: Ellipsoid matrix of earth so that (p0 + lam*dir_hat)' * B * (p0 + lam*dir_hat) = 1
        :type B: (3, 3) matrix describing earth ellipsoid in some ECEF
        """

        # Solves equation lam^2 *(d_hat'B*d_hat) + lam * 2*(p0' * B* dir_hat) + (p0' *B* p0)= 1
        a = np.transpose(dir_hat).dot(B.dot(dir_hat))
        b = 2*np.transpose(p0).dot(B.dot(dir_hat))
        c = np.transpose(p0).dot(B.dot(p0)) - 1

        p = np.zeros(3)
        p[0] = a
        p[1] = b
        p[2] = c

        lam = np.roots(p)

        hits = p0 + dir_hat*lam


        return hits 

    @staticmethod
    def calculate_sun_directions(longitude, latitude, altitude, unix_time, degrees=True):
        # Ensure all inputs are NumPy arrays
        longitude = np.asarray(longitude).flatten()
        latitude = np.asarray(latitude).flatten()
        altitude = np.asarray(altitude).flatten()
        unix_time = np.asarray(unix_time).flatten()

        n = max(longitude.size, latitude.size, altitude.size, unix_time.size)

        # If a single value is provided for any parameter, broadcast it to match the size of the other parameters
        longitude = np.broadcast_to(longitude, (n,))
        latitude = np.broadcast_to(latitude, (n,))
        altitude = np.broadcast_to(altitude, (n,))
        unix_time = np.broadcast_to(unix_time, (n,))

        phi_s = np.zeros(n)
        theta_s = np.zeros(n)

        observer = ephem.Observer()

        for i in range(n):
            # Extract a single value from the arrays
            observer.lon = str(longitude[i])
            observer.lat = str(latitude[i])
            observer.elev = altitude[i]
            
            sun = ephem.Sun()
            observer.date = datetime.utcfromtimestamp(unix_time[i])
            sun.compute(observer)

            phi_s[i] = np.rad2deg(sun.az)
            theta_s[i] = 90 - np.rad2deg(sun.alt)

        if not degrees:
            phi_s = np.deg2rad(phi_s)
            theta_s = np.deg2rad(theta_s)

        return phi_s, theta_s

    def compute_view_directions_local_tangent_plane(self):
        """Takes the intersection points and HSI camera positions and computes the angles from seabed to HSI with respect to the local tangent plane to the ellipsoid. 
        """
        n = self.rayDirectionsGlobal.shape[0]
        m = self.rayDirectionsGlobal.shape[1]

        self.seabed_to_camera_NED = np.zeros(self.camera_to_seabed_ECEF.shape)
        self.normals_ned_crs = np.zeros(self.normals_ecef_crs.shape)
        self.theta_v = np.zeros((n, m))
        self.phi_v = np.zeros((n, m))

        x_ecef = self.points_ecef_crs[:, :, 0].reshape((-1,1))
        y_ecef = self.points_ecef_crs[:, :, 1].reshape((-1,1))
        z_ecef = self.points_ecef_crs[:, :, 2].reshape((-1,1))
        
        lats, lons, alts = pm.ecef2geodetic(x = x_ecef, y = y_ecef, z = z_ecef)

        start = np.einsum('ijk, ik -> ijk', np.ones((n, m, 3), dtype=np.float64), self.position_ecef).reshape((-1,3))

        x_hsi = start[:, 0].reshape((-1,1))
        y_hsi = start[:, 1].reshape((-1,1))
        z_hsi = start[:, 2].reshape((-1,1))

        # Compute vectors from seabed intersections to HSI in NED
        NED = pm.ecef2ned(x= x_hsi, y= y_hsi, z=z_hsi, lat0 = lats, lon0=lons, h0 = alts)

        self.seabed_to_camera_NED = np.hstack((NED[0], NED[1], NED[2])).reshape((n, m, 3))

        # For remote sensing with a sun-lit seafloor:
        for i in range(n):
            
            R_ecef_2_ned = RotLib.from_matrix(rotation_matrix_ecef2ned(lon = lons[i], lat = lats[i]))

            # Calculate vector from HSI to seabed in local tangent plane NED
            #self.camera_to_seabed_NED[i, :, :] = R_ecef_2_ned.apply(self.camera_to_seabed_ECEF[i, :, :])

            # Decompose vector to angles
            polar = cartesian_to_polar(xyz = self.seabed_to_camera_NED[i,:,:])

            self.theta_v[i, :] = polar[:,1]

            self.phi_v[i, :] = polar[:,2]

            # Calculate surface normals of intersected triangles (for artificial illumination)
            self.normals_ned_crs[i, :, :] = R_ecef_2_ned.apply(self.normals_ecef_crs[i, :, :])

            


    def compute_sun_angles_local_tangent_plane(self):
        n = self.rayDirectionsGlobal.shape[0]
        m = self.rayDirectionsGlobal.shape[1]

        x_ecef = self.points_ecef_crs[:, :, 0].reshape((-1,1))
        y_ecef = self.points_ecef_crs[:, :, 1].reshape((-1,1)) 
        z_ecef = self.points_ecef_crs[:, :, 2].reshape((-1,1))

        lats, lons, alts = pm.ecef2geodetic(x = x_ecef, y = y_ecef, z = z_ecef)

        self.lats = lats
        self.lons = lons
        self.alts = alts

        phi_s, theta_s = CameraGeometry.calculate_sun_directions(longitude = lons, latitude = lats, altitude = alts, unix_time = self.unix_time, degrees = True)

        self.phi_s = phi_s.reshape((n, m, 1))

        self.theta_s = theta_s.reshape((n, m, 1))

        

    def compute_elevation_mean_sealevel(self, source_epsg, geoid_path = 'data/world/geoids/egm08_25.gtx'):
        n = self.rayDirectionsGlobal.shape[0]
        m = self.rayDirectionsGlobal.shape[1]

        x_ecef = self.position_ecef[:, 0].reshape((-1,1))
        y_ecef = self.position_ecef[:, 1].reshape((-1,1))
        z_ecef = self.position_ecef[:, 2].reshape((-1,1))

        #lats, lons, alts = pm.ecef2geodetic(x = x_ecef, y = y_ecef, z = z_ecef)

        

        alts_msl = CameraGeometry.elevation_msl(x_ecef, y_ecef, z_ecef, source_epsg=source_epsg, geoid_path = geoid_path)
        

        self.hsi_alts_msl = np.einsum('ijk, ik -> ijk', np.ones((n, m, 1), dtype=np.float32), alts_msl.reshape((-1,1)))

        
    
    @staticmethod
    def elevation_msl(x_ecef, y_ecef, z_ecef, source_epsg, geoid_path):
        """_summary_

        :param x_ecef: _description_
        :type x_ecef: _type_
        :param y_ecef: _description_
        :type y_ecef: _type_
        :param z_ecef: _description_
        :type z_ecef: _type_
        :param source_epsg: _description_
        :type source_epsg: _type_
        :param geoid_path: _description_
        :type geoid_path: _type_
        :return: _description_
        :rtype: _type_
        """
        
        with rasterio.open(geoid_path) as src:

            source_crs = CRS.from_epsg(source_epsg)
                
            target_crs = src.crs

            transformer = Transformer.from_crs(source_crs, target_crs)

            (lat, lon, alt_ell) = transformer.transform(xx=x_ecef, yy=y_ecef, zz=z_ecef)

            undulation = np.zeros(lat.shape)
            # Compute undulation and orthometric height for each point (height above MSL)
            
            for i, x, y in zip(range(lat.size), lon, lat):
                undulation[i] = next(src.sample([(float(x), float(y))]))[0]
            

        
        alt_msl = alt_ell - undulation

        return alt_msl



    

    def compute_tide_level(self, path_tide, tide_format, constant_height = 0):

        n = self.rayDirectionsGlobal.shape[0]
        m = self.rayDirectionsGlobal.shape[1]

        if path_tide == 'Undefined':

            self.hsi_tide_gridded = constant_height*np.ones((n, m, 1))

        else:

            if tide_format == 'NMA':
                try:
                    df_tide = pd.read_csv(path_tide, sep='\s+', parse_dates=[0], index_col=0, comment='#', date_parser=parser.parse)

                    # Convert the datetime index to Unix time and add column
                    df_tide['UnixTime'] = df_tide.index.astype('int64') // 10**9  # Convert nanoseconds to seconds

                    tide_height_NN2000 = 0.01*df_tide['Observations'] # Since in cm

                    tide_timestamp = df_tide['UnixTime']
                    
                    hsi_tide_interp = interp1d(x = tide_timestamp, y= tide_height_NN2000)(x = self.time)

                    # Make into gridded form:
                    self.hsi_tide_gridded = np.einsum('ijk, ik -> ijk', np.ones((n, m, 1), dtype=np.float64), hsi_tide_interp.reshape((-1,1)))
                except:
                    #print('No tide file was found!!')
                    self.hsi_tide_gridded = constant_height*np.ones((n, m, 1))



            else: # A Good place to write parsers for other formats
                TypeError
        
    def write_rgb_point_cloud(self, config, hyp, transect_string, extrapolate = True, minInd = None, maxInd = None):
        wl_red = float(config['General']['red_wave_length'])
        wl_green = float(config['General']['green_wave_length'])
        wl_blue = float(config['General']['blue_wave_length'])
        dir_point_cloud = config['Absolute Paths']['rgb_point_cloud_folder']

        wavelength_nm = np.array([wl_red, wl_green, wl_blue])

        # Localize the appropriate band indices used for analysis
        band_ind_R = np.argmin(np.abs(wavelength_nm[0] - hyp.band2Wavelength))
        band_ind_G = np.argmin(np.abs(wavelength_nm[1] - hyp.band2Wavelength))
        band_ind_B = np.argmin(np.abs(wavelength_nm[2] - hyp.band2Wavelength))

        if extrapolate == False:
            rgb = hyp.dataCubeRadiance[minInd:maxInd, :, [band_ind_R, band_ind_G, band_ind_B]]
        else:
            rgb = hyp.dataCubeRadiance[:, :, [band_ind_R, band_ind_G, band_ind_B]]


        points = self.points_ecef_crs[self.points_ecef_crs != 0].reshape((-1,3))
        rgb_points = (rgb[self.points_ecef_crs != 0] / rgb.max()).astype(np.float64).reshape((-1,3))
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(points)
        pcd.colors = o3d.utility.Vector3dVector(rgb_points)
        o3d.io.write_point_cloud(dir_point_cloud + transect_string + '.ply', pcd)


    #def transformRays(self, rays):
        # Rays are a n x m x 3 set of rays where n is each time step and coincides with the rotations and translations:

class FeatureCalibrationObject():
    def __init__(self, type, config):
        self.config = config
        self.type = type
        self.HSIPositionFeature = [] # Function of feature line
        self.HSIRotationFeature = [] #
        self.isfirst = True

    def load_cam_calibration(self, filename_cal, config):
        calHSI = CalibHSI(file_name_cal_xml=filename_cal)  # Generates a calibration object
        self.f = calHSI.f
        self.v_c = calHSI.cx
        self.k1 = calHSI.k1
        self.k2 = calHSI.k2
        self.k3 = calHSI.k3

        self.trans_x = calHSI.tx
        self.trans_y = calHSI.ty
        self.trans_z = calHSI.tz

        self.rot_x = calHSI.rx
        self.rot_y = calHSI.ry
        self.rot_z = calHSI.rz

        

        #if eval(config['General']['isFlippedRGB']):
        #    self.trans_x *= -1
        #    self.trans_y *= -1
        #if eval(config['General']['isFlippedHSI']):
        #    self.rot_z += np.pi


    def appendGeometry(self, hsiGis, cameraGeometry, binning):
        # Step 1: Define the global projected position of features
        point_feature_gt = hsiGis.features_points
        # Define the hsi pixel number
        pixel = self.bilinearInterpolation(x1_x=hsiGis.x1_x_hsi, y1_y=hsiGis.y1_y_hsi, f_Q=hsiGis.v_datacube_hsi)
        # Define the translation:
        # First define the time stamp for the four lines:
        line_nr = hsiGis.u_datacube_hsi.astype(np.int64)

        trans_Q_00 = cameraGeometry.position_ecef[line_nr[:, 0]]
        trans_Q_01 = cameraGeometry.position_ecef[line_nr[:, 1]]
        trans_Q_10 = cameraGeometry.position_ecef[line_nr[:, 2]]
        trans_Q_11 = cameraGeometry.position_ecef[line_nr[:, 3]]

        translationHSI = self.bilinearInterpolationPosition(x1_x=hsiGis.x1_x_hsi, y1_y=hsiGis.y1_y_hsi, trans_Q_00=trans_Q_00, trans_Q_01=trans_Q_01, trans_Q_10=trans_Q_10,
                                                              trans_Q_11=trans_Q_11)

        rot_Q_00 = cameraGeometry.rotation_nav_interpolated[line_nr[:, 0]]
        rot_Q_01 = cameraGeometry.rotation_nav_interpolated[line_nr[:, 1]]
        rot_Q_10 = cameraGeometry.rotation_nav_interpolated[line_nr[:, 2]]
        rot_Q_11 = cameraGeometry.rotation_nav_interpolated[line_nr[:, 3]]

        rotationRGB = self.bilinearInterpolationRotation(x1_x=hsiGis.x1_x_hsi, y1_y=hsiGis.y1_y_hsi,
                                                              rot_Q_00=rot_Q_00, rot_Q_10=rot_Q_10, rot_Q_01=rot_Q_01,
                                                              rot_Q_11=rot_Q_11)
        if self.isfirst:
            self.point_feature_gt = point_feature_gt
            self.pixel = pixel
            self.translationHSI = translationHSI
            self.rotationRGB = rotationRGB
            self.binning = np.array([binning]).reshape((1,1))
            self.diff = hsiGis.diff

            self.isfirst = False
        else:
            # Add observations
            self.diff = np.concatenate((self.diff, hsiGis.diff), axis = 0)
            self.point_feature_gt = np.concatenate((self.point_feature_gt, point_feature_gt), axis = 0)
            self.pixel = np.concatenate((self.pixel, pixel), axis = 0)
            self.translationHSI = np.concatenate((self.translationHSI, translationHSI), axis = 0)
            self.rotationRGB = np.concatenate((self.rotationRGB, rotationRGB), axis = 0)



    #def defineRayDirections
    def bilinearInterpolation(self, x1_x, y1_y, f_Q):


        f_x_y1 = (1 - x1_x)*f_Q[:, 0] + x1_x*f_Q[:, 1]
        f_x_y2 = (1 - x1_x) * f_Q[:, 2] + x1_x * f_Q[:, 3]

        f_x_y = (1 - y1_y)*f_x_y1 + y1_y*f_x_y2

        return f_x_y
    def bilinearInterpolationPosition(self, x1_x, y1_y, trans_Q_00, trans_Q_01, trans_Q_10,
                                                              trans_Q_11):

        trans_tot = np.zeros(trans_Q_00.shape)
        for i in range(3):

            f_x_y1 = (1 - x1_x)*trans_Q_00[:, i] + x1_x*trans_Q_01[:, i]
            f_x_y2 = (1 - x1_x) * trans_Q_10[:, i] + x1_x * trans_Q_11[:, i]

            f_x_y = (1 - y1_y)*f_x_y1 + y1_y*f_x_y2
            trans_tot[:, i] = f_x_y


        return trans_tot

    def bilinearInterpolationRotation(self, x1_x, y1_y, rot_Q_00, rot_Q_10, rot_Q_01, rot_Q_11):

        # Define all rotations from rot_Q_00
        delta_rot0_rot = np.zeros((4, x1_x.shape[0], 3))
        # Define all relative rotations
        delta_rot0_rot[1, :, :] = (rot_Q_01 * rot_Q_00.inv()).as_rotvec(degrees = False)
        delta_rot0_rot[2, :, :] = (rot_Q_10 * rot_Q_00.inv()).as_rotvec(degrees = False)
        delta_rot0_rot[3, :, :] = (rot_Q_11 * rot_Q_00.inv()).as_rotvec(degrees = False)

        Q_rots_permuted = np.transpose(delta_rot0_rot, axes = [2, 1, 0])

        rot_vec_final = np.zeros((3, x1_x.shape[0]))

        for i  in range(3):
            f_Q = Q_rots_permuted[i, :, :]
            f_x_y1 = (1 - x1_x) * f_Q[:, 0] + x1_x * f_Q[:, 1]
            f_x_y2 = (1 - x1_x) * f_Q[:, 2] + x1_x * f_Q[:, 3]

            f_x_y = (1 - y1_y) * f_x_y1 + y1_y * f_x_y2

            rot_vec_final[i, :] = f_x_y

        delta_rot_interp = RotLib.from_rotvec(np.transpose(rot_vec_final))

        # Compose the values to rot0
        rot_tot = delta_rot_interp * rot_Q_00


        return rot_tot

    
    # At the time of updated parameters
    def reprojectFeaturesHSI(self):
        rot_hsi_rgb = np.array([self.rot_z, self.rot_y, self.rot_x]) * 180 / np.pi

        self.rotation_hsi_rgb = RotLib.from_euler('ZYX', rot_hsi_rgb, degrees=True)

        self.rotation_hsi = self.rotationRGB * self.rotation_hsi_rgb # Composing rotations.

        self.HSIToFeaturesGlobal = self.point_feature_gt - self.translationHSI

        n = self.HSIToFeaturesGlobal.shape[0]

        self.HSIToFeaturesLocal = np.zeros(self.HSIToFeaturesGlobal.shape)
        for i in range(n):
            self.HSIToFeaturesLocal[i, :] = (self.rotation_hsi[i].inv()).apply(
                self.HSIToFeaturesGlobal[i, :])

            self.HSIToFeaturesLocal[i, :] /= self.HSIToFeaturesLocal[i, 2]




def compute_camera_rays_from_parameters(pixel_nr, cx, f, k1, k2, k3):
    """_summary_

    :param pixel_nr: _description_
    :type pixel_nr: _type_
    :param rot_x: _description_
    :type rot_x: _type_
    :param rot_y: _description_
    :type rot_y: _type_
    :param rot_z: _description_
    :type rot_z: _type_
    :param cx: _description_
    :type cx: _type_
    :param f: _description_
    :type f: _type_
    :param k1: _description_
    :type k1: _type_
    :param k2: _description_
    :type k2: _type_
    :param k3: _description_
    :type k3: _type_
    :return: _description_
    :rtype: _type_
    """

    u = pixel_nr + 1

    # Express uhi ray directions in uhi frame using line-camera model
    x_norm_lin = (u - cx) / f

    x_norm_nonlin = -(k1 * ((u - cx) / 1000) ** 5 + \
                        k2 * ((u - cx) / 1000) ** 3 + \
                        k3 * ((u - cx) / 1000) ** 2) / f

    x_norm = x_norm_lin + x_norm_nonlin

    return x_norm


def reproject_world_points_to_hsi_plane(trans_hsi_body, rot_hsi_body, pos_body, rot_body, points_world):
    """Reprojects world points to local image plane coordinates (x_h, y_h, 1)

    :param trans_hsi_body: lever arm from body center to hsi focal point
    :type trans_hsi_body: _type_
    :param rot_hsi_body: boresight rotation 
    :type rot_hsi_body: Scipy rotation object
    :param pos_body: The earth fixed earth centered position of the vehicle body centre
    :type pos_body: _type_
    :param rot_body: The rotation of the body in ECEF
    :type rot_body: Scipy rotation object
    :param points_world: World points in ECEF
    :type points_world: _type_
    :return: The reprojection coordinates
    :rtype: ndarray(n, 3)
    """

    

    # We compose the hypothesized boresight (rot_hsi_body) an lever arm (trans_hsi_body) to get the hypothesized position/orientation
    # of the hyperspectral imager
    rotation_hsi = rot_body * rot_hsi_body
    position_hsi = pos_body + rot_body.apply(trans_hsi_body)

    # Given the positions, the hsi-point vector can be expressed in ECEF
    hsi_to_feature_global = points_world - position_hsi

    # Number of features to iterate
    n = hsi_to_feature_global.shape[0]

    # We can now express the hsi-point vector in the HSI frame and normalize by the z component to find the image plane projection
    hsi_to_feature_local = np.zeros(hsi_to_feature_global.shape)
    for i in range(n):
        # The vector expressed in HSI frame
        hsi_to_feature_local[i, :] = (rotation_hsi[i].inv()).apply(hsi_to_feature_global[i, :])

        # The vector normalized to lie on the virtual plane
        hsi_to_feature_local[i, :] /= hsi_to_feature_local[i, 2]
    
    return hsi_to_feature_local




def interpolate_poses(timestamp_from, pos_from, rot_from, timestamps_to, extrapolate = True):
    """

    :param timestamp_from:
    Original timestamps
    :param pos_from: numpy array (n,3)
    Original positions
    :param rot_from: Rotation-object (n rotations)
    Original orientations
    :param timestamps_to:
    Timestamps desired for position and orientations
    :return:
    """
    """
    min_to = np.min(timestamps_to)
    max_to = np.max(timestamps_to)

    min_from = np.min(timestamp_from)
    max_from = np.max(timestamp_from)

    minInd = np.argmin(np.abs(min_from - timestamps_to))
    maxInd = np.argmin(np.abs(max_from - timestamps_to))

    # If navigation data aka "from" is made properly, min_from<<min_to, and max_from >> max_to
    
    if timestamps_to[minInd] < min_from:
        minInd += 1
    if timestamps_to[maxInd] > min_from: # So if the max index is 
        maxInd -= 1
    """

    min_ind = 0
    max_ind = timestamps_to.size


    # Setting use_absolute_position to True means that position calculations are done with absolute
    referenceGeometry = CameraGeometry(pos=pos_from,
                               rot=rot_from,
                               time=timestamp_from)

    # We exploit a method from the camera object
    referenceGeometry.interpolate(time_hsi=timestamps_to,
                          minIndRGB=min_ind,
                          maxIndRGB=max_ind,
                          extrapolate=extrapolate)



    position_to = referenceGeometry.position_nav_interpolated

    quaternion_to = referenceGeometry.rotation_nav_interpolated.as_quat()

    return position_to, quaternion_to




def cartesian_to_polar(xyz):
    """Converts from 3D cartesian coordinates to polar coordinates

    :param xyz: _description_
    :type xyz: (n,3) numpy array
    :return: Radii, Elevations, Azimuths
    :rtype: (n,3) numpy array of radii, theta, phi
    """
    polar = np.zeros(xyz.shape)
    xy = xyz[:,0]**2 + xyz[:,1]**2
    polar[:,0] = np.sqrt(xy + xyz[:,2]**2) # Radii
    polar[:,1] = np.arctan2(np.sqrt(xy), np.abs(xyz[:,2])) # for elevation angle defined from Z-axis down [0-90]
    polar[:,2] = np.arctan2(xyz[:,1], xyz[:,0]) # Azimuth

    return polar










def rotation_matrix_ecef2ned(lon, lat):
    """
    Computes the rotation matrix R from earth-fixed-earth centered (ECEF) to north-east-down (NED).
    :param lat: float in range [-90, 90]
    The latitude in degrees
    :param lon: float in range [-180, 180]
    :return R_ned_ecef: numpy array of shape (3,3)
    rotation matrix ECEF to NED
    """
    R_ned_ecef = rot_mat_ned_2_ecef(lon=lon, lat=lat)
    return np.transpose(R_ned_ecef)

def rotation_matrix_ecef2enu(lon, lat):
    l = np.deg2rad(lon)
    mu = np.deg2rad(lat)

    R_ecef_ned = rotation_matrix_ecef2ned(lon=lon, lat=lat)
    R_ecef_enu = np.zeros(R_ecef_ned.shape)
    #
    R_ecef_enu[[0, 1]] = R_ecef_ned[[1, 0]] # Swap rows
    R_ecef_enu[2] = -R_ecef_ned[2] # Switch signs to compensate for up and down


    
    return R_ecef_enu


def convert_rotation_ned_2_ecef(rot_obj_ned, position, is_geodetic = True, epsg_pos = 4326):
    """

    :param rot_obj_:
    :param position:
    The position
    :param is_geodetic: bool
    Whether position is geodetic
    :return:
    """


class GeoPose:
    """
    An object that allows to input positions with arbitrary CRS, and orientations either wrt ECEF or NED. As in other
    places in the project, the orientations are abstracted with rotation objects. The main use case is formatting poses
    to the correct CRS.
    """
    def __init__(self, timestamps, rot_obj, rot_ref, pos, pos_epsg):
        self.timestamps = timestamps
        if rot_ref == 'NED':
            self.rot_obj_ned = rot_obj
            self.rot_obj_ecef = None
        elif rot_ref == 'ECEF':
            self.rot_obj_ned = None
            self.rot_obj_ecef = rot_obj
        else:
            print('This rotation reference is not supported')
            TypeError

        # Define position
        self.position = pos
        self.epsg = pos_epsg


        self.lat = None
        self.lon = None
        self.hei = None

        self.compute_geodetic_position()
        self.compute_geocentric_orientation()
        self.compute_ned_orientation()
        self.compute_ned_2_ecef()


    def compute_geodetic_position(self, epsg_geod = 4326):
        """
            Function for transforming positions to latitude longitude height
            :param epsg_geod: int
            EPSG code of the transformed geodetic coordinate system
        """
        # If geocentric position has not been defined.
        from_CRS = CRS.from_epsg(self.epsg)
        geod_CRS = CRS.from_epsg(epsg_geod)
        transformer = Transformer.from_crs(from_CRS, geod_CRS)

        x = self.position[:, 0].reshape((-1, 1))
        y = self.position[:, 1].reshape((-1, 1))
        z = self.position[:, 2].reshape((-1, 1))

        (lat, lon, hei) = transformer.transform(xx=x, yy=y, zz=z)

        self.epsg_geod = epsg_geod
        self.lat = lat.reshape((self.position.shape[0], 1))
        self.lon = lon.reshape((self.position.shape[0], 1))
        self.hei = hei.reshape((self.position.shape[0], 1))

    def compute_geocentric_position(self, epsg_geocsc):
        """
            Function for transforming positions to geocentric
            :param epsg_geod: int
            EPSG code of the transformed geodetic coordinate system
        """

        from_CRS = CRS.from_epsg(self.epsg)
        geod_CRS = CRS.from_epsg(epsg_geocsc)
        transformer = Transformer.from_crs(from_CRS, geod_CRS)

        x = self.position[:, 0].reshape((-1, 1))
        y = self.position[:, 1].reshape((-1, 1))
        z = self.position[:, 2].reshape((-1, 1))

        (x, y, z) = transformer.transform(xx=x, yy=y, zz=z)

        self.pos_geocsc = np.concatenate((x, y, z), axis = 1)

    def compute_geocentric_orientation(self):
        if self.rot_obj_ecef == None:
            # Define rotations from ned to ecef
            if self.lat.any == None:
                # Needed to encode rotation between NED and ECEF
                self.compute_geodetic_position()

            R_body_2_ned = self.rot_obj_ned
            R_ned_2_ecef = self.compute_ned_2_ecef()
            R_body_2_ecef = R_ned_2_ecef*R_body_2_ned

            self.rot_obj_ecef = R_body_2_ecef

        else:
            pass

    def compute_ned_orientation(self):
        if self.rot_obj_ned == None:
            # Define rotations from body to ecef
            R_body_2_ecef = self.rot_obj_ecef

            # Define rotation from ecef 2 ned
            R_ecef_2_ned = self.compute_ned_2_ecef().inv()

            # Compose
            R_body_2_ned = R_ecef_2_ned*R_body_2_ecef

            self.rot_obj_ned = R_body_2_ned


        else:
            pass

    def compute_ned_2_ecef(self):

        N = self.lat.shape[0]
        rot_mats_ned_2_ecef = np.zeros((N, 3, 3), dtype=np.float64)
        
        for i in range(N):
            rot_mats_ned_2_ecef[i,:,:] = rot_mat_ned_2_ecef(lat=self.lat[i], lon = self.lon[i])


        self.rot_obj_ned_2_ecef = RotLib.from_matrix(rot_mats_ned_2_ecef)

        return self.rot_obj_ned_2_ecef

    


def rot_mat_ned_2_ecef(lat, lon):
    """
    Computes the rotation matrix R from north-east-down (NED) to earth-fixed-earth centered (ECEF)
    :param lat: float in range [-90, 90]
    The latitude in degrees
    :param lon: float in range [-180, 180]
    :return R_ned_ecef: numpy array of shape (3,3)
    rotation matrix from NED to ECEF
    """

    # Ensure lat and lon are scalar values
    lat_scalar = lat[0] if isinstance(lat, np.ndarray) else lat
    lon_scalar = lon[0] if isinstance(lon, np.ndarray) else lon

    # Convert to radians
    l = np.deg2rad(lon_scalar)
    mu = np.deg2rad(lat_scalar)

    # Compute rotation matrix
    # TODO: add source
    R_ned_ecef = np.array([[-np.cos(l) * np.sin(mu), -np.sin(l), -np.cos(l) * np.cos(mu)],
                           [-np.sin(l) * np.sin(mu), np.cos(l), -np.sin(l) * np.cos(mu)],
                           [np.cos(mu), 0, -np.sin(mu)]])
    return R_ned_ecef


def read_raster(filename, out_crs="EPSG:3857", use_z=False):
    """Read a raster to a ``pyvista.StructuredGrid``.

    This will handle coordinate transformations.
    """
    from rasterio import transform
    import rioxarray
    # Read in the data
    data = rioxarray.open_rasterio(filename)
    values = np.asarray(data)
    data.rio.nodata
    nans = values == data.rio.nodata
    if np.any(nans):
        # values = np.ma.masked_where(nans, values)
        values[nans] = np.nan
    # Make a mesh
    xx, yy = np.meshgrid(data["x"], data["y"])
    if use_z and values.shape[0] == 1:
        # will make z-comp the values in the file
        zz = values.reshape(xx.shape)
    else:
        # or this will make it flat
        zz = np.zeros_like(xx)
    mesh = pv.StructuredGrid(xx, yy, zz)
    pts = mesh.points
    lon, lat = transform(data.rio.crs, out_crs, pts[:, 0], pts[:, 1])
    mesh.points[:, 0] = lon
    mesh.points[:, 1] = lat
    mesh["data"] = values.reshape(mesh.n_points, -1, order="F")
    return mesh

def dem_2_mesh(path_dem, model_path, config):
    """
    A function for converting a specified DEM to a 3D mesh model (*.vtk, *.ply or *.stl). 
    Consequently, mesh should be thought of as 2.5D representation.
    In the case 

    :param path_dem: _description_
    :type path_dem: _type_
    :param model_path: _description_
    :type model_path: _type_
    :param config: _description_
    :type config: _type_
    """


    # The desired CRS for the model must be same as positions, orientations
    epsg_geocsc = config['Coordinate Reference Systems']['geocsc_epsg_export']

    # Intermediate point cloud format
    output_xyz = model_path.split(sep = '.')[0] + '.xyz'

    # Open the input raster dataset
    ds = gdal.Open(path_dem)

    if ds is None:
        print(f"Failed to open {path_dem}")
    else:
        # Read the first band (band index is 1)
        band = ds.GetRasterBand(1)
        no_data_value = band.GetNoDataValue()
        if band is None:
            print(f"Failed to open band 1 of {path_dem}")
        else:
            # Get the geotransform information to calculate coordinates
            geotransform = ds.GetGeoTransform()
            x_origin = geotransform[0]
            y_origin = geotransform[3]
            x_resolution = geotransform[1]
            y_resolution = geotransform[5]
            # Get the CRS information
            spatial_reference = osr.SpatialReference(ds.GetProjection())

            # Get the EPSG code
            epsg_proj = None
            if spatial_reference.IsProjected():
                epsg_proj = spatial_reference.GetAttrValue("AUTHORITY", 1)
                is_projected = True
                config.set('Coordinate Reference Systems', 'dem_epsg', str(epsg_proj))
            elif spatial_reference.IsGeographic():
                epsg_proj = spatial_reference.GetAttrValue("AUTHORITY", 0)
                proj = ds.GetProjection()
                is_projected = False


            # Automatically set
            
            
            # Get the band's data as a NumPy array of float64 (important)
            band_data = band.ReadAsArray().astype(np.float64)

            # Create a mask to identify no-data values
            mask = band_data != no_data_value

            # Create and open the output XYZ file for writing if it does not exist:
            if not os.path.exists(output_xyz):
                with open(output_xyz, 'w') as xyz_file:
                    # Write data to the XYZ file using the mask and calculated coordinates
                    for y in range(ds.RasterYSize):
                        for x in range(ds.RasterXSize):
                            if mask[y, x]:
                                x_coord = x_origin + x * x_resolution
                                y_coord = y_origin + y * y_resolution
                                xyz_file.write(f"{x_coord} {y_coord} {band_data[y, x]}\n")
            else:
                print('*.xyz already exists, ignoring re-creation')
            # Clean up
            ds = None
            band = None
    

    points = np.loadtxt(output_xyz)
    points_offset = np.mean(points, axis = 0)

    # Create a pyvista point cloud object (just to avoid precision problems)
    cloud = pv.PolyData(points-points_offset)
    
    # TODO: Apply patch to avoid crash for triangulation when using big DEM files
    # Generate a mesh from points
    mesh = cloud.delaunay_2d(progress_bar=True)

    # Transform the mesh points from projected to geocentric ECEF.
    geocsc = CRS.from_epsg(epsg_geocsc)

    if epsg_proj == 'EPSG': # If a geographic CRS
        pass
    else:
        proj = CRS.from_epsg(epsg_proj)

    transformer = Transformer.from_crs(proj, geocsc)

    # Convert to proper coordinates
    points_proj = mesh.points + points_offset

    if is_projected:
        x_proj = points_proj[:, 0].reshape((-1, 1))
        y_proj = points_proj[:, 1].reshape((-1, 1))
    else:
        x_proj = points_proj[:, 1].reshape((-1, 1))
        y_proj = points_proj[:, 0].reshape((-1, 1))
    
    h_proj = points_proj[:, 2].reshape((-1, 1))

    (x_ecef, y_ecef, z_ecef) = transformer.transform(xx=x_proj, yy=y_proj, zz=h_proj)

    # Introduce mesh offsets in ECEF sized numbers seem to cause troubles:
    offset_x = np.mean(x_ecef.reshape(-1))
    offset_y = np.mean(y_ecef.reshape(-1))
    offset_z = np.mean(z_ecef.reshape(-1))

    mesh.points[:, 0] = x_ecef.reshape(-1) - offset_x
    mesh.points[:, 1] = y_ecef.reshape(-1) - offset_y
    mesh.points[:, 2] = z_ecef.reshape(-1) - offset_z
   

    # Save mesh
    mesh.save(model_path)
    # Save mesh meta
    # Define your metadata dictionary
    metadata = {
        "offset_x": offset_x,
        "offset_y": offset_y,
        "offset_z": offset_z,
        "epsg_code": geocsc.to_epsg(),  # Example EPSG code, replace with your actual code
        "data_type": str(mesh.points.dtype),  # Add other metadata entries here
    }

    model_meta_path = model_path.split('.')[0] + '_meta.json' 
    # Open the file in write mode with proper indentation
    with open(model_meta_path, "w") as f:
        # Write the dictionary to the file using JSON dump
        json.dump(metadata, f)


def crop_geoid_to_pose(path_dem, config, geoid_path = 'data/world/geoids/egm08_25.gtx'):
    # The desired CRS for the model must be same as positions, orientations
    epsg_geocsc = config['Coordinate Reference Systems']['geocsc_epsg_export']

    # Open the input raster dataset
    
    ds = gdal.Open(geoid_path)

    df_pose = pd.read_csv(config['Absolute Paths']['pose_path'])

    # Find CRS of DEM
    spatial_reference = osr.SpatialReference(ds.GetProjection())


    # Get the EPSG code
    epsg_proj = None
    if spatial_reference.IsProjected():
        epsg_proj = spatial_reference.GetAttrValue("AUTHORITY", 1)
    elif spatial_reference.IsGeographic():
        epsg_proj = spatial_reference.GetAttrValue("AUTHORITY", 0)
    
    # Transform points to DEM CRS
    geocsc = CRS.from_epsg(epsg_geocsc)
    proj = ds.GetProjection()
    transformer = Transformer.from_crs(geocsc, proj)

    x_ecef = df_pose[' X'].values.reshape((-1, 1))
    y_ecef = df_pose[' Y'].values.reshape((-1, 1))
    z_ecef = df_pose[' Z'].values.reshape((-1, 1))

    (x_proj, y_proj, z_proj) = transformer.transform(xx=x_ecef, yy=y_ecef, zz=z_ecef)

    coords_horz = np.concatenate((x_proj.reshape((-1,1)), y_proj.reshape((-1,1))), axis = 1)

    # Determine bounding rectangle ()
    polygon = MultiPoint(coords_horz).envelope
    # Corners
    x, y = polygon.exterior.xy
    x = np.array(x)
    y = np.array(y)

    # Determine padding
    padding = float(config['General']['max_ray_length'])

    if spatial_reference.IsProjected():
        # Add padding to 
        xmin = x.min() - padding
        ymin = y.min() - padding
        xmax = x.max() + padding
        ymax = y.max() + padding
    elif spatial_reference.IsGeographic():
        # Must translate metric padding into increments in lon/lat
        (x_new, y_new, z_new) = pm.enu2geodetic(e= np.array([-padding, padding]), n= np.array([-padding, padding]), u = 0, lon0=np.mean(y), lat0=np.mean(x), h0 = 0)
        delta_x = x_new[1] - np.mean(x)
        delta_y = y_new[1] - np.mean(y)

        # Swap x and y because x, y above is latitude, longitude and Rasterio expects opposite
        ymin = x.min() - delta_x
        xmin = y.min() - delta_y
        ymax = x.max() + delta_x
        xmax = y.max() + delta_y

        minx, miny, maxx, maxy = [xmin, ymin, xmax, ymax]  # Replace with actual coordinates

        

    # Crops the DEM to the appropriate bounds and writes a new file (Copies CRS info from Geoid)
    crop_dem_from_bounds(minx, miny, maxx, maxy, dem_path_source=geoid_path, dem_path_target=path_dem)
     


def crop_dem_from_bounds(minx, miny, maxx, maxy, dem_path_source, dem_path_target):

    dem_dataset = rasterio.open(dem_path_source)
    res_x = dem_dataset.transform.a
    res_y = -dem_dataset.transform.e

    if maxx-minx < res_x:
        minx += -res_x
        maxx += res_x
    
    if maxy-miny < res_y:
        miny += -res_y
        maxy += res_y

    window = from_bounds(minx, miny, maxx, maxy, dem_dataset.transform)

    window = Window(window.col_off, window.row_off, np.ceil(window.width), np.ceil(window.height))

    cropped_dem_data = dem_dataset.read(window=window)

    # Update the transform based on the new window
    new_transform = dem_dataset.window_transform(window)
    new_height, new_width = cropped_dem_data.shape[1], cropped_dem_data.shape[2]

    # Create a new dataset for the cropped DEM
    cropped_dem_profile = dem_dataset.profile.copy()
    cropped_dem_profile.update({
        'width': new_width,
        'height': new_height,
        'transform': new_transform
    })


    with rasterio.open(dem_path_target, 'w', **cropped_dem_profile) as dst:
        dst.write(cropped_dem_data)


def position_transform_ecef_2_llh(position_ecef, epsg_from, epsg_to, config):
    """
    Function for transforming ECEF positions to latitude longitude height
    :param position_ecef: numpy array floats (n,3)
    :param epsg_from: int
    EPSG code of the original geocentric coordinate system (ECEF)
    :param epsg_to: int
    EPSG code of the transformed geodetic coordinate system
    :return lat_lon_hei: numpy array floats (n,3)
    latitude, longitude ellipsoid height.
    """
    geocsc = CRS.from_epsg(epsg_from)
    geod = CRS.from_epsg(epsg_to)
    transformer = Transformer.from_crs(geocsc, geod)

    x_ecef = position_ecef[:, 0].reshape((-1, 1))
    y_ecef = position_ecef[:, 1].reshape((-1, 1))
    z_ecef = position_ecef[:, 2].reshape((-1, 1))

    (lat, lon, hei) = transformer.transform(xx=x_ecef, yy=y_ecef, zz=z_ecef)

    lat_lon_hei = np.zeros(position_ecef.shape)
    lat_lon_hei[:, 0] = lat.reshape((position_ecef.shape[0], 1))
    lat_lon_hei[:, 1] = lon.reshape((position_ecef.shape[0], 1))
    lat_lon_hei[:, 2] = hei.reshape((position_ecef.shape[0], 1))

    return lat_lon_hei



import pyvista as pv
from pyvistaqt import BackgroundPlotter
import pandas as pd
import numpy as np
from scipy.spatial.transform import Rotation
import open3d as o3d
import pymap3d as pm
# A simple visualization of various types of data
from gref4hsi.utils.geometry_utils import rotation_matrix_ecef2ned, rotation_matrix_ecef2enu

def show_mesh_camera(config, show_mesh = True, show_pose = True, ref_frame = 'ECEF'):
    """
    # Reads the mesh file and pose info, and plots the trajectory next to the mesh.

    :param config: Dictionary-like (configuration) object
    :return: Nothing

    """


    # Todo: encode show mesh camera to use h5-embedded data? Or is this a loss of performance?
    mesh_path = config['Absolute Paths']['model_path']
    try:
        texture_path = config['Absolute Paths']['tex_path']
    except KeyError:
        texture_path = None
        
    pose_path = config['Absolute Paths']['pose_path']


    pose = pd.read_csv(
        pose_path, sep=',',
        header=0)

    points_cam_ecef = np.concatenate( (pose[" X"].values.reshape((-1,1)), pose[" Y"].values.reshape((-1,1)), pose[" Z"].values.reshape((-1,1))), axis = 1)
    use_local = False
    if use_local == True:
        eul_cam = np.concatenate((pose[" Yaw"].values.reshape((-1, 1)),
                                  pose[" Pitch"].values.reshape((-1, 1)),
                                  pose[" Roll"].values.reshape((-1, 1))), axis=1)
    else:
        eul_cam = np.concatenate( (pose[" RotZ"].values.reshape((-1,1)), pose[" RotY"].values.reshape((-1,1)), pose[" RotX"].values.reshape((-1,1))), axis = 1)


    R_body_to_ecef = Rotation.from_euler("ZYX", eul_cam, degrees=True)
    

    # Read the mesh
    mesh = pv.read(mesh_path)
    if ref_frame == 'NED':
        x = np.mean(points_mesh_ecef[:,0])
        y = np.mean(points_mesh_ecef[:,1])
        z = np.mean(points_mesh_ecef[:,2])
        
        lat0, lon0, hei0 = pm.ecef2geodetic(x, y, z, deg=True)
        R_ecef_to_ned = Rotation.from_matrix(rotation_matrix_ecef2ned(lon=lon0, lat=lat0))

        R_body_to_ned = R_ecef_to_ned*R_body_to_ecef
        rotMats = R_body_to_ned.as_matrix()

        points_mesh_ecef = mesh.points

        # Next part is to make a NED
        x_cam, y_cam, z_cam = pm.ecef2ned(x = points_cam_ecef[:,0], y = points_cam_ecef[:,1], z = points_cam_ecef[:,2], lon0=lon0, lat0=lat0, h0=hei0)

        points_cam = np.concatenate((x_cam.reshape((-1,1)), y_cam.reshape((-1,1)), z_cam.reshape((-1,1))), axis = 1)

        x_mesh, y_mesh, z_mesh = pm.ecef2ned(x = points_mesh_ecef[:,0], y = points_mesh_ecef[:,1], z = points_mesh_ecef[:,2], lon0=lon0, lat0=lat0, h0=hei0)

        points_mesh = np.concatenate((x_mesh.reshape((-1,1)), y_mesh.reshape((-1,1)), z_mesh.reshape((-1,1))), axis = 1)

        mesh.points = points_mesh


    elif ref_frame == 'ENU':
        lat0, lon0, hei0 = pm.ecef2geodetic(x, y, z, deg=True)
        R_ecef_to_enu = Rotation.from_matrix(rotation_matrix_ecef2enu(lon=lon0, lat=lat0))

        R_body_to_enu = R_ecef_to_enu*R_body_to_ecef
        rotMats = R_body_to_enu.as_matrix()

        points_mesh_ecef = mesh.points

        # Next part is to make a NED
        x_cam, y_cam, z_cam = pm.ecef2enu(x = points_cam_ecef[:,0], y = points_cam_ecef[:,1], z = points_cam_ecef[:,2], lon0=lon0, lat0=lat0, h0=hei0)

        points_cam = np.concatenate((x_cam.reshape((-1,1)), y_cam.reshape((-1,1)), z_cam.reshape((-1,1))), axis = 1)

        x_mesh, y_mesh, z_mesh = pm.ecef2enu(x = points_mesh_ecef[:,0], y = points_mesh_ecef[:,1], z = points_mesh_ecef[:,2], lon0=lon0, lat0=lat0, h0=hei0)

        points_mesh = np.concatenate((x_mesh.reshape((-1,1)), y_mesh.reshape((-1,1)), z_mesh.reshape((-1,1))), axis = 1)
        mesh.points = points_mesh
    
    else:
        rotMats = R_body_to_ecef.as_matrix()
        points_cam = points_cam_ecef


    

    p = BackgroundPlotter(window_size=(600, 400))
    
    if show_mesh:
        if texture_path != None:
            tex = pv.read_texture(texture_path)
            p.add_mesh(mesh, texture=tex)
        else:
            p.add_mesh(mesh)

    if show_pose:
        p.add_points(points_cam, render_points_as_spheres=True,
                        point_size=1)
        directionX = rotMats[:, :, 0]
        directionY = rotMats[:, :, 1]
        directionZ = rotMats[:, :, 2]
        step = 10
        scale = np.linalg.norm(np.max(points_cam, axis = 0)-np.min(points_cam, axis = 0), axis=0)
        p.add_arrows(points_cam[::step], directionX[::step], mag = 0.1*scale, color = 'red')
        p.add_arrows(points_cam[::step], directionY[::step], mag=0.1*scale, color= 'green')
        p.add_arrows(points_cam[::step], directionZ[::step], mag=0.1*scale, color= 'blue')
    p.show()
    p.app.exec_()

def show_camera_geometry(CameraGeometry, config):
    mesh_path = config['General']['model_path']
    texture_path = config['General']['tex_path']




    rotMats = CameraGeometry.rotation_hsi.as_matrix()

    points_cam = CameraGeometry.position_ecef

    mesh = pv.read(mesh_path)
    tex = pv.read_texture(texture_path)

    p = BackgroundPlotter(window_size=(600, 400))

    p.add_mesh(mesh, texture=tex)
    p.add_points(points_cam, render_points_as_spheres=True,
                      point_size=1)
    directionX = rotMats[:, :, 0]
    directionY = rotMats[:, :, 1]
    directionZ = rotMats[:, :, 2]
    step = 10
    p.add_arrows(points_cam[::step], directionX[::step], mag = 0.25, color= 'red')
    p.add_arrows(points_cam[::step], directionY[::step], mag=0.25, color= 'green')
    p.add_arrows(points_cam[::step], directionZ[::step], mag=0.25, color= 'blue')


    # Add direction vectors in cyan to show FOV
    direction1 = CameraGeometry.rayDirectionsGlobal[:, 0, :]
    direction2 = CameraGeometry.rayDirectionsGlobal[:, -1, :]

    p.add_arrows(points_cam[::step], direction1[::step], mag=1, color='cyan')
    p.add_arrows(points_cam[::step], direction2[::step], mag=1, color='yellow')

    p.show()
    p.app.exec_()

def show_projected_hsi_points(HSICameraGeometry, config, transect_string):
    mesh_path = config['Absolute Paths']['model_path']
    
    point_cloud_path = config['Absolute Paths']['rgb_point_cloud_folder'] + transect_string + '.ply'

    rotMats = HSICameraGeometry.rotation_hsi.as_matrix()
    points_cam = HSICameraGeometry.position_ecef

    mesh = pv.read(mesh_path)

    try:
        texture_path = config['Absolute Paths']['tex_path']
        tex = pv.read_texture(texture_path)
    except:
        pass
    p = BackgroundPlotter(window_size=(600, 400))

    p.add_mesh(mesh)
    p.add_points(points_cam, render_points_as_spheres=True,
                 point_size=2, color = 'black')
    directionX = rotMats[:, :, 0]
    directionY = rotMats[:, :, 1]
    directionZ = rotMats[:, :, 2]
    step = 100
    p.add_arrows(points_cam[::step], directionX[::step], mag=0.25, color='red')
    p.add_arrows(points_cam[::step], directionY[::step], mag=0.25, color='green')
    p.add_arrows(points_cam[::step], directionZ[::step], mag=0.25, color='blue')

    # Add the points from intersections
    #points_intersection = HSICameraGeometry.projection[::step, ::step, :].reshape((-1,3))

    #p.add_points(points_intersection, render_points_as_spheres=True,
    #             point_size=1)


    pcd = o3d.io.read_point_cloud(point_cloud_path)
    color_arr = np.asarray(pcd.colors)
    hyp_pcl = pv.read(point_cloud_path)
    hyp_pcl['colors'] = color_arr
    p.add_mesh(hyp_pcl, scalars='colors', rgb=True, point_size=2)

    p.show()
    p.app.exec_()

def show_point_clouds(pathPcl1, pathPcl2, ind1 = None, ind2 = None, hyp1 = None, hyp2 = None):

    if ind1 == None:
        p = BackgroundPlotter(window_size=(600, 400))
        pcd = o3d.io.read_point_cloud(pathPcl1)
        color_arr = np.asarray(pcd.colors)
        p.set_background(color='white', top=None)
        #color_arr[:, 0] = (color_arr[:, 0] - color_arr[:, 0].min()) / (color_arr[:, 0].max() - color_arr[:, 0].min())
        #color_arr[:, 1] = (color_arr[:, 1] - color_arr[:, 1].min()) / (color_arr[:, 1].max() - color_arr[:, 1].min())
        #color_arr[:, 2] = (color_arr[:, 2] - color_arr[:, 2].min()) / (color_arr[:, 2].max() - color_arr[:, 2].min())
        hyp_pcl = pv.read(pathPcl1)
        hyp_pcl['colors'] = color_arr
        p.add_mesh(hyp_pcl, scalars='colors', rgb=True, point_size=2)

        pcd = o3d.io.read_point_cloud(pathPcl2)
        color_arr = np.asarray(pcd.colors)


        #color_arr[:, 0] = (color_arr[:, 0] - color_arr[:, 0].min()) / (color_arr[:, 0].max() - color_arr[:, 0].min())
        #color_arr[:, 1] = (color_arr[:, 1] - color_arr[:, 1].min()) / (color_arr[:, 1].max() - color_arr[:, 1].min())
        #color_arr[:, 2] = (color_arr[:, 2] - color_arr[:, 2].min()) / (color_arr[:, 2].max() - color_arr[:, 2].min())

        hyp_pcl2 = pv.read(pathPcl2)
        hyp_pcl2['colors'] = color_arr
        p.add_mesh(hyp_pcl2, scalars='colors', rgb=True, point_size=2)

        p.add_points(hyp1.position_hsi, render_points_as_spheres=True,
                     point_size=1)
        p.add_points(hyp2.position_hsi, render_points_as_spheres=True,
                     point_size=1)

        p.show()
        p.app.exec_()
    else:
        p = BackgroundPlotter(window_size=(600, 400))
        pcd = o3d.io.read_point_cloud(pathPcl1)
        color_arr = np.asarray(pcd.colors)
        hyp_pcl = pv.read(pathPcl1)
        hyp_pcl['colors'] = color_arr
        p.add_mesh(hyp_pcl, scalars='colors', rgb=True, point_size=2)

        pcd = o3d.io.read_point_cloud(pathPcl2)
        color_arr = np.asarray(pcd.colors)
        hyp_pcl2 = pv.read(pathPcl2)
        hyp_pcl2['colors'] = color_arr
        p.add_mesh(hyp_pcl2, scalars='colors', rgb=True, point_size=2)

        p.show()
        p.app.exec_()




def main():
    show_mesh_camera()




if __name__ == '__main__':
    main()





import time

import numpy as np
import pyvista as pv
from pykdtree.kdtree import KDTree
from pyvista import examples
import trimesh

# Record the start time
start_time = time.time()

# Load data
data = examples.load_random_hills()
data.translate((10, 10, 10))




# Create triangular plane (vertices [10, 0, 0], [0, 10, 0], [0, 0, 10])
size = 10
vertices = np.array([[size, 0, 0], [0, size, 0], [0, 0, size]])
face = np.array([3, 0, 1, 2])

planes = pv.PolyData(vertices, face)

# Subdivide plane so we have multiple points to project to
planes = planes.subdivide(8)

# Get origins and normals
origins = planes.cell_centers().points
normals = planes.compute_normals(cell_normals=True, point_normals=False)["Normals"]

# Convert faces to nx3 format
faces = data.faces.reshape(-1, 3)

# Convert PolyData to trimesh.Trimesh
data_mesh = trimesh.Trimesh(vertices=data.points, faces=faces)

# Vectorized Ray trace
"""points, pt_inds, cell_inds = data.multi_ray_trace(
    origins, normals
)"""  # Must have rtree, trimesh, and pyembree installed


ray_mesh_intersector = trimesh.ray.ray_pyembree.RayMeshIntersector(geometry=data_mesh)

# Perform ray tracing using trimesh.ray.intersects_id
cell_inds, pt_inds, points = ray_mesh_intersector.intersects_id(ray_origins=origins,  
                                                                ray_directions=normals, 
                                                                multiple_hits=False,
                                                                return_locations=True)


# Filter based on distance threshold, if desired (mimics VTK ray_trace behavior)
# threshold = 10  # Some threshold distance
# distances = np.linalg.norm(origins[inds] - points, ord=2, axis=1)
# inds = inds[distances <= threshold]

tree = KDTree(data.points.astype(np.double))
_, data_inds = tree.query(points)

elevations = data.point_data["Elevation"][data_inds]

# Mask points on planes
planes.cell_data["Elevation"] = np.zeros((planes.n_cells,))
planes.cell_data["Elevation"][pt_inds] = elevations
planes.set_active_scalars("Elevation")  # Probably not necessary, but just in case

# Create axes
axis_length = 20
tip_length = 0.25 / axis_length * 3
tip_radius = 0.1 / axis_length * 3
shaft_radius = 0.05 / axis_length * 3
x_axis = pv.Arrow(
    direction=(axis_length, 0, 0),
    tip_length=tip_length,
    tip_radius=tip_radius,
    shaft_radius=shaft_radius,
    scale="auto",
)
y_axis = pv.Arrow(
    direction=(0, axis_length, 0),
    tip_length=tip_length,
    tip_radius=tip_radius,
    shaft_radius=shaft_radius,
    scale="auto",
)
z_axis = pv.Arrow(
    direction=(0, 0, axis_length),
    tip_length=tip_length,
    tip_radius=tip_radius,
    shaft_radius=shaft_radius,
    scale="auto",
)
x_label = pv.PolyData([axis_length, 0, 0])
y_label = pv.PolyData([0, axis_length, 0])
z_label = pv.PolyData([0, 0, axis_length])
x_label.point_data["label"] = [
    "x",
]
y_label.point_data["label"] = [
    "y",
]
z_label.point_data["label"] = [
    "z",
]


# Calculate the elapsed time
elapsed_time = time.time() - start_time

# Print the elapsed time in seconds
print(f"Elapsed Time: {elapsed_time:.2f} seconds")

if elapsed_time < 2:
    print('Test run took less than 2 seconds ==> Test passed')
else:
    print('Test run took more than 2 seconds ==> Test failed')
# Plot results
p = pv.Plotter()
p.add_mesh(x_axis, color="r")
p.add_point_labels(x_label, "label", show_points=False, font_size=24)
p.add_mesh(y_axis, color="r")
p.add_point_labels(y_label, "label", show_points=False, font_size=24)
p.add_mesh(z_axis, color="r")
p.add_point_labels(z_label, "label", show_points=False, font_size=24)
p.add_mesh(data)
p.add_mesh(planes)
p.show()



import numpy as np

import cv2 as cv

class Image():
    def __init__(self, array):
        """Takes in an image array with nxmx3 of type uint8"""
        self.image_array = array

    def clahe_adjustment(self, is_luma = False):
        tile_size = 8
        clip_lim = 3
        clahe = cv.createCLAHE(clipLimit=clip_lim, tileGridSize=(tile_size, tile_size))
        if is_luma:
            img = (self.luma_array*255).astype(np.uint8) # 0->1

            # Apply equalization and adjust luma array
            self.luma_array = clahe.apply(img).reshape((img.shape[0], img.shape[1], 1))

        else:
            
            img = self.image_array
            R = clahe.apply(img[:, :, 0]).reshape((img.shape[0], img.shape[1], 1))
            G = clahe.apply(img[:, :, 1]).reshape((img.shape[0], img.shape[1], 1))
            B = clahe.apply(img[:, :, 2]).reshape((img.shape[0], img.shape[1], 1))

            self.clahe_adjusted = np.concatenate((R, G, B), axis=2)

    def to_luma(self, gamma, image_array, gamma_inverse = False, gamma_value = 0.45):
        """Takes in an image array with nxmx3 of type uint8 and transforms it to luminance with or without gamma compensation"""
        # Image is transformed into sRGB luminance
        R = (image_array[:, :, 2] / 255)

        G = (image_array[:, :, 1] / 255)

        B = (image_array[:, :, 0] / 255)

        # It is possible to apply a Gamma Conversion to the results as examplified below
        if gamma:
            R[R <= 0.04045] = R[R < 0.04045] / 12.92
            R[R > 0.04045] = ((R[R > 0.04045] + 0.055) / 1.055) ** 2.4
            G[G <= 0.04045] = G[G < 0.04045] / 12.92
            G[G > 0.04045] = ((G[G > 0.04045] + 0.055) / 1.055) ** 2.4
            B[B <= 0.04045] = B[B < 0.04045] / 12.92
            B[B > 0.04045] = ((B[B > 0.04045] + 0.055) / 1.055) ** 2.4

        # Linear approach for calculating CIE luminance rec. 709, ref Eq. (7).
        self.luma_array =  0.2125 * R + 0.7154 * G + 0.0721 * B

        if gamma_inverse:
            self.luma_array = np.power(self.luma_array, gamma_value).clip(0,1)




# Python built-ins
import sys
import os
import configparser
from os import path

# Third party libraries
import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation as RotLib
from scipy.spatial.transform import Rotation
import h5py
from scipy.spatial.transform import Rotation as RotLib
import pandas as pd
import pymap3d as pm
from scipy.spatial.transform import Slerp
from scipy.interpolate import interp1d
from pyproj import CRS, Transformer

# Local modules
from gref4hsi.utils.geometry_utils import CameraGeometry, GeoPose
from gref4hsi.utils.geometry_utils import rot_mat_ned_2_ecef, interpolate_poses
from gref4hsi.utils.geometry_utils import dem_2_mesh, crop_geoid_to_pose


class Hyperspectral:
    """
    Class for storing/accessing acquired hyperspectral data. One instance corresponds to a transect chunk (*.h5).
    """
    def __init__(self, filename, config, load_datacube = True):
        """
        Instantiate transect chunk object from chunk file name and config file.
        :param filename: string
        path to *.h5 file
        :param config: config type
        dictionary-like interface to configuration file
        """
        # Chunk filename
        self.name = filename
        # If errors arise due to improper dataset paths below, open the *.h5 file in hdfviewer and edit the path.
        with h5py.File(filename, 'r', libver='latest') as self.f:
            # TODO: Embed the paths from this special format into confic file
            try:
                dataCubePath = config['HDF.hyperspectral']['dataCube']
            except:
                pass
            
            try:
                timestampHyperspectralPath = config['HDF.hyperspectral']['timestamp']
                self.dataCubeTimeStamps = self.f[timestampHyperspectralPath][()]
            except:
                pass

            try:
                # The band center wavelengths
                band2WavelengthPath = config['HDF.calibration']['band2Wavelength']
                self.band2Wavelength = self.f[band2WavelengthPath][()]
            except:
                pass
            
            try:
                # The band widths (if they exist)
                fwhmPath = config['HDF.calibration']['fwhm']

                self.fwhm = self.f[fwhmPath][()]
            except:
                pass

            try:
                # These should come together in the case that radiometric calibration should be done
                radiometricFramePath = config['HDF.calibration']['radiometricFrame']
                darkFramePath = config['HDF.calibration']['darkFrame']
                exposureTimePath = config['HDF.hyperspectral']['exposureTime']
                is_uhi = config['HDF']['is_uhi']

                
                if eval(is_uhi):
                    self.t_exp = self.f[exposureTimePath][()][0] / 1000  # Recorded in milliseconds
                else:
                    self.t_exp = self.f[exposureTimePath][()] / 1000

                self.darkFrame = self.f[darkFramePath][()]
                self.radiometricFrame = self.f[radiometricFramePath][()]
            except:
                pass
            
            try:
                RGBFramesPath = config['HDF.rgb']['rgbFrames']
                timestampRGBPath = config['HDF.rgb']['timestamp']
                try:
                    # Extract RGB image data if available
                    self.RGBTimeStamps = self.f[timestampRGBPath][()]
                    self.RGBImgs = self.f[RGBFramesPath][()]
                    self.n_imgs = self.RGBTimeStamps.shape[0]
                except (UnboundLocalError, KeyError):
                    pass
            except:
                pass
            
            

            if load_datacube:
                # To save memory, we made it an option to not load datacube
                self.dataCube = self.f[dataCubePath][()]

                self.n_scanlines = self.dataCube.shape[0]
                self.n_pix = self.dataCube.shape[1]
                self.n_bands = self.dataCube.shape[1]

                # Calculate radiance cube if this is cube is not calibrated
                self.digital_counts_2_radiance(config=config)

            # Check if the dataset exists
            processed_nav_folder = config['HDF.processed_nav']['folder']
            
            # Read the position da
            if processed_nav_folder in self.f:
                try:
                    processed_nav_config = config['HDF.processed_nav']
                    self.pos_ref = self.f[processed_nav_config['position_ecef']][()]
                    self.quat_ref = self.f[processed_nav_config['quaternion_ecef']][()]
                    self.pose_time = self.f[processed_nav_config['timestamp']][()]
                except:
                    pass
            



       

            



    def digital_counts_2_radiance(self, config):

        # Only calibrate if data it is not already done
        is_calibrated = eval(config['HDF.hyperspectral']['is_calibrated'])

        if is_calibrated == is_calibrated:
            self.dataCubeRadiance = self.dataCube.astype(np.float32)

            # Add the radiance dataset
            radiance_cube_path = config['HDF.hyperspectral']['dataCube']
        else:
            self.dataCubeRadiance = np.zeros(self.dataCube.shape, dtype = np.float32)
            for i in range(self.dataCube.shape[0]):
                self.dataCubeRadiance[i, :, :] = (self.dataCube[i, :, :] - self.darkFrame) / (
                        self.radiometricFrame * self.t_exp)
            
            # Add the radiance dataset
            radiance_cube_path = config['HDF.hyperspectral']['dataCube'] + '_radiance'
            
            # TODO: Write the radiance data to the h5 file
            Hyperspectral.add_dataset(data = self.dataCubeRadiance, name=radiance_cube_path, h5_filename=self.name)
        
        config.set('HDF.hyperspectral', 'radiance_cube', radiance_cube_path)

        # For memory efficiency
        del self.dataCube

        
            
            
    
    """def add_dataset(self, data, name):
        
        # The h5 file structure can be studied by unravelling the structure in Python or by using HDFview
        with h5py.File(self.name, 'a', libver='latest') as self.f:
            # Check if the dataset exists
            if name in self.f:
                del self.f[name]

            dset = self.f.create_dataset(name=name, data = data)
    """
    
    @staticmethod
    def add_dataset(data, name, h5_filename):
        """
        Method to write a dataset to the h5 file
        :param data: type any permitted (see h5py doc)
        The data to be written
        :param name: string
        The path/name of the dataset
        :param h5_filename: string
        The path to the h5_file
        :return: None
        """
        # The h5 file structure can be studied by unravelling the structure in Python or by using HDFview
        with h5py.File(h5_filename, 'a', libver='latest') as f:
            # Check if the dataset exists
            if name in f:
                del f[name]

            dset = f.create_dataset(name=name, data = data)

    """def get_dataset(self, dataset_name):
        
        Method to return a dataset by the name
        :param dataset_name: string
        The h5 dataset path
        :return: dataset
        Returns the queried dataset. Could be many datatypes, but mainly numpy arrays in our usage.
        
        # The h5 file structure can be studied by unravelling the structure in Python or by using HDFview
        with h5py.File(self.name, 'a', libver='latest') as self.f:
            dataset = self.f[dataset_name][()]
        return dataset"""
    @staticmethod
    def get_dataset(h5_filename, dataset_name):
        """equivalent to the method, except it does not need an instance of a Hyperspectral object

        :param h5_filename: Path to h5/hdf file
        :type h5_filename: string
        :param dataset_name: h5 rooted path, e.g. processed/reflectance/remote_sensing_reflectance
        :type dataset_name: string
        :return: The dataset at the relevant location
        :rtype: numpy array or other
        """
        # The h5 file structure can be studied by unravelling the structure in Python or by using HDFview
        with h5py.File(h5_filename, 'a', libver='latest') as f:
            dataset = f[dataset_name][()]
        return dataset

class DataLogger:
    def __init__(self, filename, header):
        self.filename = filename
        # Generate file with header
        with open(filename, 'w') as fh:
            fh.write(header + '\n')

    def append_data(self, data):
        # Append data line to file
        with open(self.filename, 'a') as fh:
            # Generate data line to file
            if data[0] != None:
                line = ','.join([str(el) for el in data])

                # Append line to file
                fh.write(line + '\n')

def ardupilot_extract_pose(config, iniPath):
    
    """
        :param config: The processing configuration object (from */configuration.ini file)
        :param iniPath: The path to */configuration.ini file
        :return: 
    """
    
    # Attitude retrieved from an ardupilot CSV
    att_path = config['General']['ardupath'] + 'Att.csv'
    df_att = pd.read_csv(att_path)

    # Convert the DataFrame to a matrix (list of lists)
    matrix_att = np.array(df_att.values.tolist())
    
    # Rotations in roll-pitch-yaw define the rotation of the vehicle's body frame (i.e. IMU)
    Rotation = RotLib.from_euler("ZYX", np.flip(matrix_att[:,1:4], axis = 1), degrees=True)
    time_rot = matrix_att[:,0]
    
    # Position retrieved from an ardupilot CSV
    pos_path = config['General']['ardupath'] + 'pos.csv'
    df_pos = pd.read_csv(pos_path)
    matrix_pos = np.array(df_pos.values.tolist())

    time_pos = matrix_pos[:, 0]
    Position = matrix_pos[:, 1:4]

    linearPositionInterpolator = interp1d(time_pos, np.transpose(Position), fill_value='extrapolate')
    position_nav_interpolated = np.transpose(linearPositionInterpolator(time_rot))

    pose_path = config['General']['ardupath'] + 'pose.csv'

    config.set('General', 'pose_path', pose_path)

    epsg_proj = 4326 # Standard
    epsg_geocsc = config['General']['modelepsg']
    # Transform the mesh points to ECEF.

    geocsc = CRS.from_epsg(epsg_geocsc)
    proj = CRS.from_epsg(epsg_proj)
    transformer = Transformer.from_crs(proj, geocsc)

    points_proj = position_nav_interpolated

    lat = points_proj[:, 0].reshape((-1, 1))
    lon = points_proj[:, 1].reshape((-1, 1))
    hei = points_proj[:, 2].reshape((-1, 1))

    (x_ecef, y_ecef, z_ecef) = transformer.transform(xx=lat, yy=lon, zz=hei)

    pos_geocsc = np.concatenate((x_ecef.reshape((-1,1)), y_ecef.reshape((-1,1)), z_ecef.reshape((-1,1))), axis = 1)

    #dlogr = DataLogger(pose_path, 'CameraLabel, X, Y, Z, Roll, Pitch, Yaw, RotX, RotY, RotZ')



    data_matrix = np.zeros((len(time_rot), 10))
    for i in range(len(time_rot)):

        # Ardupilot delivers geodetic positions (EPSG::
        pos_geod = position_nav_interpolated[i, :]

        # Define rotation matrix between local tangential plane (NED, North-East-Down) and Earth-Centered-Earth-Fixed (ECEF).
        R_n_e = rot_mat_ned_2_ecef(lat=pos_geod[0], lon=pos_geod[1])

        Rotation_n = RotLib.from_matrix(R_n_e)

        Rotation_e = Rotation_n*Rotation[i] # Composing rotations

        roll = matrix_att[i, 1]
        pitch = matrix_att[i, 2]
        yaw = matrix_att[i, 3]

        

        rotz, roty, rotx = Rotation_e.as_euler("ZYX", degrees=True)

        pos = pos_geocsc[i,:]

        data_matrix[i,:] = np.array([time_rot[i], pos[0], pos[1], pos[2], roll, pitch, yaw, rotx, roty, rotz])

        #dlogr.append_data([time_rot[i], pos[0], pos[1], pos[2], roll, pitch, yaw, rotx, roty, rotz])



    headers = ['Timestamp', ' X', ' Y', ' Z', ' Roll', ' Pitch', ' Yaw', ' RotX', ' RotY', ' RotZ']

    # Create a DataFrame from the data_matrix and headers
    df = pd.DataFrame(data_matrix, columns=headers)

    # Save the DataFrame as a CSV file
    df.to_csv(pose_path, index=False)

    with open(iniPath, 'w') as configfile:
        config.write(configfile)
        
    return config


def reformat_h5_embedded_data_h5(config, config_file):
    """
                Parses pose from h5, interpolates and reformats. Writes the positions and orientations of hyperspectral frames to
                    dataset under the /Nav folder

                Args:
                    config_file: The *.ini file containing the paths and configurations.

                Returns:
                    None: The function returns nothing.
    """


    # Traverse through h5 dir to append the data to file
    h5_folder = config['Absolute Paths']['h5_folder']
    is_first = True
    for filename in sorted(os.listdir(h5_folder)):
        
        # Find the interesting prefixes
        if filename.endswith('h5') or filename.endswith('hdf'):
            # Identify the total path
            path_hdf = h5_folder + filename

            # Read out the data of a file
            hyp = Hyperspectral(path_hdf, config, load_datacube = False)


            # If rotation reference is quaternion:
            # Alternatives quat, eul_ZYX
            rotation_reference_type = config['HDF.raw_nav']['rotation_reference_type']
            if rotation_reference_type == 'quat':
                quaternion_ref = hyp.get_dataset(h5_filename=path_hdf, dataset_name=config['HDF.raw_nav']['quaternion'])
                if quaternion_ref.shape[0] == 4:
                    quaternion_ref = np.transpose(quaternion_ref)
                
            
                quaternion_convention  = config['HDF.raw_nav']['quaternion_convention']

                # If scalar-first convention
                if quaternion_convention == 'wxyz':
                    quat_temp = quaternion_ref
                    quaternion_ref = np.roll(quaternion_ref, shift=-1, axis=1)
                elif quaternion_convention == 'xyzw':
                    # We use the xyzw convention for scipy, see:
                    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.from_quat.html
                    pass
                else:
                    raise ValueError("The quaternion convention must be defined")
            elif rotation_reference_type == 'eul_ZYX':
                eul_ref = hyp.get_dataset(h5_filename=path_hdf,dataset_name=config['HDF.raw_nav']['eul_ZYX'])
                if eul_ref.shape[0] == 3:
                    eul_ref = np.transpose(eul_ref)
                eul_is_degrees = eval(config['HDF.raw_nav']['eul_is_degrees'])
                
                ## Assuming It is given as roll, pitch, yaw
                eul_ZYX_ref = np.flip(eul_ref, axis = 1)
                quaternion_ref = RotLib.from_euler(angles = eul_ZYX_ref, seq = 'ZYX', degrees = eul_is_degrees).as_quat()
                
            else:
                raise ValueError("The rotation reference type must be defined as quat or eul_ZYX")


            is_global_rot = eval(config['HDF.raw_nav']['is_global_rot'])

            # Parse position
            position_ref = hyp.get_dataset(h5_filename=path_hdf,dataset_name=config['HDF.raw_nav']['position'])
            
            if position_ref.shape[0] == 3:
                position_ref = np.transpose(position_ref)


            timestamps_imu = hyp.get_dataset(h5_filename=path_hdf, dataset_name=config['HDF.raw_nav']['timestamp'])\
                .reshape(-1)


            # The rotation of the reference
            rot_obj = RotLib.from_quat(quaternion_ref)

            timestamp_hsi = hyp.dataCubeTimeStamps.reshape(-1)

            # Compute interpolated absolute positions positions and orientation:
            position_interpolated, quaternion_interpolated = interpolate_poses(timestamp_from=timestamps_imu,
                                                             pos_from=position_ref,
                                                             rot_from=rot_obj,
                                                             timestamps_to=timestamp_hsi)

            # If the original orientations are with respect to (North-East-Down) NED
            if not is_global_rot:
                rot_ref = 'NED'
            else:
                rot_ref = 'ECEF'

            # The positions supplied for the reference
            pos_epsg_orig = config['Coordinate Reference Systems']['pos_epsg_orig']

            # The positions supplied for exporting the model
            pos_epsg_export = config['Coordinate Reference Systems']['geocsc_epsg_export']

            # The interpolated rotation-object
            rot_interpolated = RotLib.from_quat(quaternion_interpolated)

            # For simple geo-pose for changing between formats.
            geo_pose_ref = GeoPose(timestamps=timestamp_hsi,
                                   rot_obj= rot_interpolated, 
                                   rot_ref=rot_ref,
                                   pos = position_interpolated, 
                                   pos_epsg=pos_epsg_orig)

            # Convert position to the epsg used for the 3D model
            geo_pose_ref.compute_geocentric_position(epsg_geocsc=pos_epsg_export)


            # Calculate the geodetic position using the WGS-84 (for conversion of orientations)
            epsg_wgs84 = 4326
            geo_pose_ref.compute_geodetic_position(epsg_geod=epsg_wgs84)

            # Calculate ECEF orientation
            geo_pose_ref.compute_geocentric_orientation()

            # For readability, extract the position and orientation wrt ECEF
            position_ref_ecef = geo_pose_ref.pos_geocsc
            quaternion_ref_ecef = geo_pose_ref.rot_obj_ecef.as_quat()

            # For stacking in CSV
            rot_vec_ecef = geo_pose_ref.rot_obj_ecef.as_euler('ZYX', degrees=True)

            partial_pose = np.concatenate((timestamp_hsi.reshape((-1,1)), position_ref_ecef, rot_vec_ecef), axis=1)
            if is_first:
                # Initialize
                total_pose = partial_pose
                # Ensure that flag is not raised again. Offsets should only be set once.
                is_first = False
            else:
                # Concatenate/stack poses
                total_pose = np.concatenate((total_pose, partial_pose), axis=0)
                is_first = False
            
            

            # Add camera position
            position_ref_name = config['HDF.processed_nav']['position_ecef']
            Hyperspectral.add_dataset(data=position_ref_ecef, name=position_ref_name, h5_filename=path_hdf)

            # Add camera quaternion
            quaternion_ref_name = config['HDF.processed_nav']['quaternion_ecef']
            Hyperspectral.add_dataset(data=quaternion_ref_ecef, name=quaternion_ref_name, h5_filename=path_hdf)

            # Add time stamps
            time_stamps_name = config['HDF.processed_nav']['timestamp']
            Hyperspectral.add_dataset(data=timestamp_hsi, name=time_stamps_name, h5_filename=path_hdf)


            


    headers = ['Timestamp', ' X', ' Y', ' Z', ' RotZ', ' RotY', ' RotX']

    # Create a DataFrame from the data_matrix and headers
    # Stores the entire pose path
    df = pd.DataFrame(total_pose, columns=headers)

    pose_path = config['Absolute Paths']['pose_path']

    try:
        # Save the DataFrame as a CSV file
        df.to_csv(pose_path, index=False)
    except:
        # Extract directory
        directory = os.path.dirname(pose_path)

        # Check if directory doesn't exist and create it
        if not os.path.exists(directory):
            os.makedirs(directory)
        
        df.to_csv(pose_path, index=False)

# The main script for aquiring pose.        
def export_pose(config_file):
    """
        Parses pose and embeds it into the *.h5 file. Writes the positions and orientations of hyperspectral frames to
            dataset under the /Nav folder

        Args:
            config_file: The *.ini file containing the paths and configurations.

        Returns:
            config: The function returns config object to allow that it is modified.
    """
    config = configparser.ConfigParser()
    config.read(config_file)

    # This file handles pose exports of various types
    # As of now there are three types:

    # 1) Agisoft, h5_embedded and independent_file
    try:
        pose_export_type = config['General']['pose_export_type']
    except:
        # Default to h5_embedded
        pose_export_type = 'h5_embedded'

    if pose_export_type == 'h5_embedded':
        reformat_h5_embedded_data_h5(config=config,
                                              config_file=config_file)
    elif pose_export_type == 'independent_file':
        print('There is no support for this export functionality! Fix immediately!')
        config = -1
    else:
        print('File type: ' + pose_export_type + 'type is not defined!')
        config = -1

    
    
def export_model(config_file):
    """"""
    config = configparser.ConfigParser()
    config.read(config_file)

    # This file handles model exports of various types
    # As of now there are three types:
    # 1) Agisoft, *.ply file/DEM and None
    model_export_type = config['General']['model_export_type']

    if model_export_type == 'agisoft':
        agisoft_export_model(config_file=config_file)
    elif model_export_type == 'ply_file':
        # Nothing needs to be done then?
        pass
    elif model_export_type == 'dem_file':
        file_path_dem = config['Absolute Paths']['dem_path']
        file_path_3d_model = config['Absolute Paths']['model_path']
        # Make new only once.

        if os.path.exists(file_path_3d_model):
            print('3D model already exists and overwriting is not supported')
            pass
        else:
            dem_2_mesh(path_dem=file_path_dem, model_path=file_path_3d_model, config=config)

    elif model_export_type == 'geoid':
        file_path_dem = config['Absolute Paths']['dem_path']
        file_path_geoid = config['Absolute Paths']['geoid_path']
        file_path_3d_model = config['Absolute Paths']['model_path']

        # Crop the DEM to appropriate size based on poses and maximum ray length
        crop_geoid_to_pose(path_dem=file_path_dem, config=config, geoid_path=file_path_geoid)

        # Make into a 3D model
        dem_2_mesh(path_dem=file_path_dem, model_path=file_path_3d_model, config=config)







if __name__ == "__main__":
    # Here we could set up necessary steps on a high level. 
    args = sys.argv[1:]
    config_file = args[0]



from collections import namedtuple
import configparser
import os
import sys

import numpy as np

# Detect OS and set FPs
if os.name == 'nt':
    # Windows OS
    base_fp = 'D:'
    home = 'C:/Users/haavasl'
elif os.name == 'posix':
    # This Unix-like systems inl. Mac and Linux
    base_fp = '/media/haavasl/Expansion'
    home = '/home/haavasl'


# Use this if working with the github repo to do quick changes to the module
module_path = os.path.join(home, 'VsCodeProjects/gref4hsi/')
if module_path not in sys.path:
    sys.path.append(module_path)

from gref4hsi.utils import parsing_utils, uhi_parsing_utils
from gref4hsi.scripts import georeference, orthorectification
from gref4hsi.utils import visualize
from gref4hsi.utils.config_utils import prepend_data_dir_to_relative_paths, customize_config

# D:\HyperspectralDataAll\UHI\2020-07-01-14-40-15-ArcticSeaIce-Ben-Lange
DATA_DIR = os.path.join(base_fp, "HyperspectralDataAll/UHI/2020-07-01-14-40-15-ArcticSeaIce-Ben-Lange/")

# The configuration file stores the settings for georeferencing
config_file_mission = os.path.join(DATA_DIR, 'configuration.ini')


config_path_template = os.path.join(home, 'VsCodeProjects/gref4hsi/data/config_examples/configuration_uhi.ini')

# Copies the template to config_file_mission and sets up the necessary directories
prepend_data_dir_to_relative_paths(config_path=config_path_template, DATA_DIR=DATA_DIR)

# Non-default settings
custom_config = {'General':
                    {'mission_dir': DATA_DIR,
                    'model_export_type': 'dem_file', # Infer seafloor structure from altimeter recordings
                    'max_ray_length': 5,
                    'lab_cal_dir': os.path.join(base_fp, 'HyperspectralDataAll/UHI/Lab_Calibration_Data/NP')}, # Max distance in meters from UHI to seafloor

                'Coordinate Reference Systems': 
                    {'proj_epsg' : 3395, # The projected CRS for orthorectified data (an arctic CRS)
                    'geocsc_epsg_export' : 4978, # 3D cartesian system for earth consistent with GPS frame (but inconsistent with eurasian techtonic plate)
                    'dem_epsg' : 3395, # (Optional) If you have a DEM this can be used
                    'pos_epsg_orig' : 4978}, # The CRS of the positioning data we deliver to the georeferencing

                'Relative Paths':
                    {'dem_folder': 'Input/GIS/'}, # Using altimeter, we generate one DEM per transect chunk
                
                'Absolute Paths':
                    {'geoid_path': os.path.join(home, "VsCodeProjects/gref4hsi/data/world/geoids/egm08_25.gtx")}, # Using altimeter, we generate one DEM per transect chunk

                'Orthorectification':
                    {'resample_rgb_only': True, # Good choice for speed
                    'resolutionhyperspectralmosaic': 0.01, # 1 cm
                    'raster_transform_method': 'north_east'},
                
                'HDF.raw_nav': {'altitude': 'raw/nav/altitude',
                    'rotation_reference_type' : 'eul_ZYX', # The vehicles orientations are used as Yaw, Pitch, Roll
                    'is_global_rot' : False, # The vehicles orientations are used as Yaw, Pitch, Roll
                    'eul_is_degrees' : True},

                'HDF.calibration': {'band2wavelength' : 'processed/radiance/calibration/spectral/band2Wavelength',
                                'darkframe' : 'processed/radiance/calibration/radiometric/darkFrame',
                                'radiometricframe' : 'processed/radiance/calibration/radiometric/radiometricFrame',
                                'fov' : 'processed/radiance/calibration/geometric/fieldOfView'},
                
                # Where to find the standard data for the cube. Note that is_calibrated implies whether data is already in correct format
                'HDF.hyperspectral': {'datacube' : 'processed/radiance/dataCube',
                                    'exposuretime' : 'processed/radiance/exposureTime',
                                    'timestamp' : 'processed/radiance/timestamp',
                                    'is_calibrated' : True},

                'HDF.rgb' :{'rgb_frames' : 'rawdata/rgb/rgbFrames',
                            'rgb_frames_timestamp' : 'rawdata/rgb/timestamp'}
                
}

# Customizes the config file
customize_config(config_path=config_file_mission, dict_custom=custom_config)

# Settings specific to the pre-processing of UHI data. At present they are hardcoded, but they could be integrated 
SettingsPreprocess = namedtuple('SettingsPreprocessing', ['dtype_datacube', 
                                                            'rotation_matrix_hsi_to_body',
                                                            'translation_body_to_hsi',
                                                            'rotation_matrix_alt_to_body',
                                                            'translation_alt_to_body',
                                                            'config_file_name',
                                                            'time_offset_sec',
                                                            'lon_lat_alt_origin',
                                                            'resolution_dem',
                                                            'agisoft_process'])

config_uhi_preprocess = SettingsPreprocess(dtype_datacube = np.float32,
                            # The fill value for empty cells (select values not occcuring in cube or ancillary data)
                            rotation_matrix_hsi_to_body = np.array([[0, 1, 0],
                                                                    [1, 0, 0],
                                                                    [0, 0, -1]]),
                            # Boolean being expressing whether to rectify only composite (true) or data cube and composite (false). True is fast.
                            translation_body_to_hsi = np.array([0, 0, 0]),
                            rotation_matrix_alt_to_body = np.array([[0, 1, 0],
                                                                    [1, 0, 0],
                                                                    [0, 0, -1]]),
                            # Boolean being expressing whether to rectify only composite (true) or data cube and composite (false). True is fast.
                            translation_alt_to_body = np.array([0.5, 0, 0]),
                            time_offset_sec =  22,
                            # Ben's tick s1 starts at 1593614097.992003 -> 22 s delay
                            # Ben's tick s2 starts at 1593614414.995001 -> 0 s delay

                            lon_lat_alt_origin =  np.array([1, 1, 0]),
                            # The beast sets up a fake coordinate system at 1 deg lon/lat.
                            config_file_name = 'configuration.ini',
                            resolution_dem = 0.2, # The resolution of the Altimeter-based DEM
                            agisoft_process  = False) # This is an option for photogrammetry-based processing in the case you have imagery





"""config = configparser.ConfigParser()
config.read(config_file)"""

def main():

    # The config stores all relevant configurations
    config = configparser.ConfigParser()
    config.read(config_file_mission)

    # The minimum for georeferencing is to parse 1) Mesh model and 2) The pose of the reference
    uhi_parsing_utils.uhi_beast(config=config, config_uhi=config_uhi_preprocess)
    
    config = parsing_utils.export_pose(config_file_mission)

    config = configparser.ConfigParser()

    config.read(config_file_mission)
    
    # Exports model
    parsing_utils.export_model(config_file_mission)

    # Visualize the data 3D photo model from RGB images and the time-resolved positions/orientations
    #visualize.show_mesh_camera(config)

    # Georeference the line scans of the hyperspectral imager. Utilizes parsed data
    georeference.main(config_file_mission, viz=False)

    orthorectification.main(config_file_mission)
    # Alternatively mode = 'calibrate'
    # georeference.main(config_file, mode='calibrate', is_calibrated=True)


if __name__ == "__main__":
    main()


import configparser
import os


def prepend_data_dir_to_relative_paths(config_path, DATA_DIR, mkdirs = True):
    """The config file holds a series of default relative paths. By providing a parent directory (one per mission)
    the absolute paths can be established. To automatically generate the folder structure (make the directories), allow mkdirs=True

    :param config_path: path to config file for read and write
    :type config_path: str
    :param DATA_DIR: path to data directory. The modified config is put here.
    :type DATA_DIR: str
    :param mkdirs: Whether to make folders, defaults to True.
    :type mkdirs: bool, optional
    """
    config = configparser.ConfigParser()
    config.read(config_path)

    # Set the value
    config.set('General', 'mission_dir', DATA_DIR)

    if mkdirs:
        if os.path.exists(DATA_DIR):
                pass
        else:
            os.makedirs(DATA_DIR, exist_ok=True)

    # Check if the 'Absolute Paths' section already exists
    if 'Absolute Paths' not in config:
        # Create a new section for 'Absolute Paths'
        config.add_section('Absolute Paths')

    # Iterate through the key-value pairs in 'Relative Paths'
    for key, relative_path in config['Relative Paths'].items():
        # Copy the key-value pair to 'Absolute Paths' section
        absolute_path = os.path.join(DATA_DIR, relative_path)
        config.set('Absolute Paths', key, absolute_path)

        # Only directories end with separators
        is_dir = (len(absolute_path.split('.'))==1)

        # Only create directories if path is a valid directory
        if mkdirs and is_dir:
            # If it exists, do nothing
            if os.path.exists(absolute_path):
                pass
            else:
                os.makedirs(absolute_path, exist_ok=True)

    # Save the updated configuration to the file
    
    config_path_write = os.path.join(DATA_DIR, 'configuration.ini')
    
    with open(config_path_write, 'w') as configfile:
        config.write(configfile)

def customize_config(config_path, dict_custom):
    """Allows changing dictionary-specified entries of the config to custom values

    :param config_path: path to config file for read and write
    :type config_path: str
    :param dict_custom: dictionary of custom options for processing.
    :type dict_custom: nested dictionary. Dictionary of sections which are dictionaries
    """
    config = configparser.ConfigParser()
    config.read(config_path)

    

    # dict_custom like config is a two-level nested dictionary
    for key_section, section_dict in dict_custom.items():
        # Iterate through each section, and assign custom values 
        for key_sub_section, value_custom in section_dict.items():
            try:
                config.set(key_section, key_sub_section, str(value_custom))
            except configparser.NoSectionError:
                config.add_section(key_section)
                config.set(key_section, key_sub_section, str(value_custom))



    # Save the updated configuration to the file
    with open(config_path, 'w') as configfile:
        config.write(configfile)

